# -*- coding: utf-8 -*-
# Copyright (C) 2008  Warp Networks, S.L.
# Author:  Pablo Martí
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


from time import time

import dbus
from dbus.mainloop.glib import DBusGMainLoop
from dbus.service import Object, BusName, method, signal
gloop = DBusGMainLoop(set_as_default=True)

from twisted.application.service import Service
from twisted.internet import defer, reactor
from twisted.python import log

from wader.cli.wrapper import DBusWrapper
from wader.common.profiles import DBusProfileManager
from wader.common.consts import WADER_SERVICE, WADER_OBJPATH, WADER_INTFACE
from wader.common.dialer import DialerManager
from wader.common.hardware import DBusExporterHelper, HardwareManager
from wader.common.startup import attach_serial_protocol

import wader.common.exceptions as ex

# states
IDLE, ADDING, REMOVING = range(3)

ADD_THRESHOLD = 7.

class CoreSerialService(Service):
    def __init__(self):
        self.ctrl = None
        self.prof = None
        self.dial = None

    def startService(self):
        self.ctrl = CoreStartupController()
        self.prof = DBusProfileManager()
        self.dial = DialerManager(self.ctrl)


class CoreStartupController(Object, DBusExporterHelper):
    """
    I handle devices in the system

    Discovery, identification, hotplugging, etc.

    @ivar clients: Dictionary with a reference to every configured device
    """
    def __init__(self):
        name = BusName(WADER_SERVICE, bus=dbus.SessionBus(mainloop=gloop))
        super(CoreStartupController, self).__init__(bus_name=name,
                                                    object_path=WADER_OBJPATH)

        self.hm = HardwareManager()
        # configured clients
        self.clients = {}
        # list of potential clients
        self.potential_clients = []
        # list with all the udis added during hotplugging
        self.added_udis = []
        # time for next action
        self.last_action = None
        self.call_id = None
        self.mode = IDLE

        self._populate_potential_clients()
        self._connect_to_signals()

    def _connect_to_signals(self):
        for s, m in [('DeviceAdded', self._device_added_cb),
                     ('DeviceRemoved', self._device_removed_cb)]:
            self.hm.manager.connect_to_signal(s, m)

    def _populate_potential_clients(self):
        parent_udis = self.hm._get_parent_udis()
        self.potential_clients.extend(parent_udis)

    def _device_added_cb(self, udi):
        if self.mode == IDLE:
            self.mode = ADDING

        self.last_action = time()
        self.added_udis.append(udi)
        try:
            if not self.call_id:
                self.call_id = reactor.callLater(ADD_THRESHOLD,
                                             self._process_added_udis)
            else:
                self.call_id.reset(ADD_THRESHOLD)
        except:
            # call has already been fired
            self.added_udis = []
            self.call_id = None
            self.mode = IDLE

    def _device_removed_cb(self, udi):
        if self.mode == ADDING:
            # we're in the middle of a hotpluggin event and the udis that
            # we just added to self.added_udis are disappearing!
            # whats going on? Some devices such as the Huawei E870 will
            # add some child udis, and will remove them once libusual kicks
            # in, so we need to wait for at most ADD_THRESHOLD seconds
            # since the last removal/add to find out what really got added
            if udi in self.added_udis:
                self.added_udis.remove(udi)

        if udi in self.clients:
            log.msg("Device %s removed, deleting client instance" % udi)
            self.DeviceRemoved(udi)
            self._remove_client(udi)

        elif udi in self.potential_clients:
            self.DeviceRemoved(udi)
            self.potential_clients.remove(udi)

    def _process_found_devices(self, devices):
        for udi in [dev.udi for dev in devices]:
            self.DeviceAdded(udi)

    def _process_added_udis(self):
        assert self.mode == ADDING
        # obtain the parent udis of all the devices with modem capabilities
        parent_udis = self.hm._get_parent_udis()
        # we're only interested on devices not being handled and just added
        not_handled_udis = set(self.clients.keys()) ^ parent_udis
        just_added_udis = not_handled_udis & set(self.added_udis)
        # get the devices out of the UDIs and emit DeviceAdded for each udi
        d = self.hm._get_devices_from_udis(just_added_udis)
        d.addCallback(self._process_found_devices)
        # cleanup
        self.mode = IDLE
        self.added_udis = []
        self.call_id = None

    def _attach_device(self, plugin, udi):
        device = attach_serial_protocol(plugin, udi)
        statemachine_callbacks = statemachine_errbacks = {}
        device.wrapper = DBusWrapper(device, {},
                            statemachine_callbacks,
                            statemachine_errbacks)
        self._register_client(device, udi)
        return device.udi

    def _remove_client(self, udi):
        """
        Removes client identified by C{udi} from C{self.clients}
        """
        log.msg("Deleting client %s" % udi)
        d = defer.maybeDeferred(self.clients[udi].close)
        def really_remove(ignored):
            del self.clients[udi]
        d.addCallback(really_remove)
        return d

    def _register_client(self, plugin, udi):
        """
        Registers C{plugin} in {self.clients} by its C{udi}
        """
        log.msg("Registering %s for %s" % (plugin, udi))
        self.clients[udi] = plugin

    def _get_devices(self):
        parent_udis = self.hm._get_parent_udis()
        not_handled_udis = set(self.clients.keys()) ^ parent_udis
        return self.hm._get_devices_from_udis(parent_udis)

    @method(WADER_INTFACE, in_signature='', out_signature='ao',
           async_callbacks=('async_cb', 'async_eb'))
    def GetDevices(self, async_cb, async_eb):
        """
        Returns a list of object paths with all the found devices

        It also includes the object paths of already handled devices
        """
        d = self._get_devices()
        d.addCallback(lambda devices: [dev.udi for dev in devices])
        return self.add_generic_callbacks(d, async_cb, async_eb)

    @method(WADER_INTFACE, in_signature='s', out_signature='o',
           async_callbacks=('async_cb', 'async_eb'))
    def AddDevice(self, udi, async_cb, async_eb):
        """
        Configure and leave ready to use the device identified by C{udi}
        """
        log.msg("AddDevice called with udi %s" % udi)
        if udi in self.clients:
            raise ex.DeviceAlreadyAddedError("Device %s is already handled" % udi)

        plugin = self.hm.get_device_from_udi(udi)

        if not plugin:
            raise dbus.DBusException("No device found for udi %s" % udi)

        from wader.common.hardware.base import identify_device
        d = identify_device(plugin)
        d.addCallback(self._attach_device, udi)
        return self.add_generic_callbacks(d, async_cb, async_eb)

    @method(WADER_INTFACE, in_signature='o', out_signature='',
           async_callbacks=('async_cb', 'async_eb'))
    def RemoveDevice(self, udi, async_cb, async_eb):
        """
        Free and remove the device identified by C{udi}
        """
        log.msg("RemoveDevice called with udi %s" % udi)

        if not udi in self.clients:
            raise dbus.DBusException("Device %s not handled" % udi)

        d = self._remove_client(udi)
        return self.add_generic_callbacks_and_swallow(d, async_cb, async_eb)

    @signal(WADER_INTFACE, signature='o')
    def DeviceAdded(self, udi):
        """
        Emitted when a 3G device is added
        """
        log.msg("Emitting DeviceAdded %s" % udi)
        if udi not in self.potential_clients:
            self.potential_clients.append(udi)
        else:
            log.err("DeviceAdded calld with udi already in potential_clients")

    @signal(WADER_INTFACE, signature='o')
    def DeviceRemoved(self, udi):
        """
        Emitted when a 3G device is removed
        """
        log.msg("Emitting DeviceRemoved %s" % udi)

