# -*- coding: utf-8 -*-
# Copyright (C) 2006-2008  Vodafone España, S.A.
# Copyright (C) 2008  Warp Networks, S.L.
# Author:  Pablo Martí
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Config related classes and methods

The object L{config} is a singleton instance of WaderConfig. The overhead of
opening the object, performing a get/set and closing it was too much.
That's why its a singleton. The instance is closed during
L{wader.common.shutdown.shutdown_core()}. You must be careful as importing it
too early during the first run of the program, could cause that the config
file is not created yet and you'd get a messy traceback. Delay its import as
much as possible in modules used at startup or that are imported at startup.
"""

from ConfigParser import ConfigParser
import os
import pickle
import shutil
import tempfile

import dbus
from dbus.service import Object, BusName, method

from wader.common.consts import PROFTEMPLATE, CONFTEMPLATE, \
                                WADER_SERVICE, PRF_INTFACE

class WaderConfigBase(object):
    """I manage Wader's configuration"""

    def __init__(self, path):
        self.path = path
        self.fileobj = None
        self.conf = ConfigParser()
        self._read_settings()

    def _read_settings(self):
        self.fileobj = open(self.path)
        # Read defaults
        self.conf.read([CONFTEMPLATE])
        # Read user configuration
        self.conf.readfp(self.fileobj)

    def get(self, section, option):
        return self.conf.get(section, option)

    def getboolean(self, section, option):
        return self.conf.getboolean(section, option)

    def getint(self, section, option):
        return self.conf.getint(section, option)

    def write(self):
        """
        Saves the current configuration to L{self.path}
        """
        return self.conf.write(open(self.path, 'w'))

    def set(self, section, option, value):
        return self.conf.set(section, option, value)

    def setboolean(self, section, option, val):
        if isinstance(val, str):
            if val not in ['yes', 'no']:
                raise AttributeError("Invalid boolean value %s" % val)

            value = val
        else:
            if not isinstance(val, bool):
                raise AttributeError("What I'm supposed to do with %?" % val)

            value = val and 'yes' or 'no'

        return self.conf.set(section, option, value)

    def serialize(self, path=None):
        """
        Returns the path where the profile has been serialized

        The caller is responsible of removing the file
        """
        if not path:
            path = tempfile.mkstemp('', 'WaderConfig', '/tmp', text=True)[1]
        fobj = open(path, 'w')
        pickle.dump(self, fobj, pickle.HIGHEST_PROTOCOL)
        fobj.close()
        return path

    def close(self):
        self.fileobj.close()
        self.conf = None


class MobileProfile(Object, WaderConfigBase):
    """
    I am a mobile profile on the system
    """

    def __init__(self, opath, profpath):
        WaderConfigBase.__init__(self, profpath)
        busname = BusName(WADER_SERVICE, bus=dbus.SessionBus())
        Object.__init__(self, busname, object_path=opath)
        self.opath = opath

    def __repr__(self):
        return '<MobileProfile at %s>' % self.opath

    def __eq__(self, other):
        if not isinstance(other, (MobileProfile, dict)):
            raise ValueError("Imposible to compare me with %s" % other)

        if isinstance(other, dict):
            for key, val in self.conf.items('connection'):
                if other[key] != val:
                    return False

            return True
        else:
            for key, val in self.conf.items('connection'):
                if other.get('connection', key) != val:
                    return False

            return True

    def __ne__(self, other):
        if not isinstance(other, (dict, MobileProfile)):
            raise ValueError("Imposible to compare me with %s" % other)

        return not self.__eq__(other)

    def patch(self, other):
        if isinstance(other, dict):
            for k, v in other.iteritems():
                self.set('connection', k, v)
        else:
            for key, val in other.conf.items('connection'):
                self.set('connection', key, val)

        self.write()

    @method(PRF_INTFACE, in_signature='s', out_signature='s')
    def Get(self, key):
        return self.get('connection', key)

    @method(PRF_INTFACE, in_signature='', out_signature='a(ss)')
    def Items(self):
        return [(k, v) for k, v in self.conf.items('connection')]

    @classmethod
    def from_path(cls, opath, profpath):
        if not os.path.exists(profpath):
            shutil.copy(PROFTEMPLATE, profpath)

        return cls(opath, profpath)

    @classmethod
    def from_dict(cls, opts, opath, profpath):
        profile = MobileProfile.from_path(opath, profpath)

        for k, v in opts.iteritems():
            profile.set('connection', k, v)

        profile.write()
        return profile

    def close(self):
        WaderConfigBase.close(self)
        self.remove_from_connection()
