# -*- coding: utf-8 -*-
# Copyright (C) 2006-2008  Vodafone España, S.A.
# Copyright (C) 2008  Warp Networks, S.L.
# Author:  Pablo Martí
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Daemons for Wader
"""



from twisted.internet.task import LoopingCall
from twisted.python import log

from wader.common.netspeed import NetworkSpeed
import wader.common.exceptions as ex
import wader.common.notifications as N

class WaderDaemon(object):
    """
    I represent a Daemon in Wader

    A Daemon is an entity that performs a repetitive action, like polling
    signal quality from the datacard. A Daemon will emit fake notifications
    built out of the action's response.
    """
    def __init__(self, frequency, device):
        super(WaderDaemon, self).__init__()
        self.frequency = frequency
        self.device = device
        self.loop = None

    def __repr__(self):
        return self.__class__.__name__

    def start(self):
        """Starts the Daemon"""
        log.msg("DAEMONS: DAEMON %s started..." % self.__class__.__name__)
        if not self.loop or not self.loop.running:
            self.loop = LoopingCall(self.function)
            self.loop.start(self.frequency)

            args = (self.function, self.frequency)
            log.msg("DAEMONS: executing %s every %d seconds" % args)

    def stop(self):
        """Stops the Daemon"""
        if self.loop.running:
            cname = self.__class__.__name__
            log.msg("DAEMONS: DAEMON %s stopped..." % cname)
            self.loop.stop()

    def function(self):
        """Function that will be called periodically"""
        raise NotImplementedError()


class SignalQualityDaemon(WaderDaemon):
    """
    I emit fake SIG_RSSI UnsolicitedNotifications
    """

    def function(self):
        d = self.device.sconn.get_signal_level()
        d.addCallback(lambda rssi: self.device.sconn.RSSIChanged(rssi))


class CellTypeDaemon(WaderDaemon):
    """
    I emit fake SIG_NEW_CONN_MODE UnsolicitedNotifications
    """

    def function(self):
        def get_network_info_eb(failure):
            failure.trap(ex.NetworkTemporalyUnavailableError)
            log.err(failure, "CellTypeDaemon: FAILURE RECEIVED")

        d = self.device.sconn.get_network_info()
        d.addCallback(lambda netinfo:
                      self.device.sconn.CellTypeChanged(netinfo[1]))
        d.addErrback(get_network_info_eb)


class NetworkSpeedDaemon(WaderDaemon):
    """
    I emit fake SIG_SPEED UnsolicitedNotifications
    """
    def __init__(self, frequency, device):
        super(NetworkSpeedDaemon, self).__init__(frequency, device)
        self.netspeed = NetworkSpeed()

    def start(self):
        log.msg("DAEMONS: DAEMON %s started..." % self)
        self.netspeed.start()
        self.loop = LoopingCall(self.function)
        self.loop.start(self.frequency)

    def stop(self):
        log.msg("DAEMONS: DAEMON %s stopped..." % self)

        self.loop.stop()
        self.netspeed.stop()

    def function(self):
        up, down = self.netspeed['up'], self.netspeed['down']
        self.device.sconn.SpeedChanged(up, down)

class WaderDaemonCollection(object):
    """
    I am a collection of Daemons

    I provide some methods to manage the collection.
    """
    def __init__(self):
        self.daemons = {}

    def append_daemon(self, name, daemon):
        """Adds C{daemon} to the collection with C{name}"""
        self.daemons[name] = daemon

    def has_daemon(self, name):
        """Returns True if C{name} exists"""
        return name in self.daemons

    def remove_daemon(self, name):
        """Removes daemon with C{name}"""
        try:
            del self.daemons[name]
        except KeyError:
            raise

    def start_daemons(self, arg=None):
        """Starts all daemons"""
        for daemon in self.daemons.values():
            daemon.start()

    def stop_daemon(self, name):
        """Stops daemon with C{name}"""
        try:
            self.daemons[name].stop()
        except KeyError:
            raise

    def stop_daemons(self):
        """Stops all daemons"""
        for daemon in self.daemons.values():
            daemon.stop()


class WaderDaemonFactory(object):
    """Daemon Factory for Wader"""

    @classmethod
    def build_daemon_collection(cls, device):
        """
        Returns a C{WaderServiceCollection} customized for C{device}
        """
        col = WaderDaemonCollection()

        if device.has_two_ports():
            # check capabilities
            if N.SIG_RSSI not in device.custom.device_capabilities:
                # device doesn't sends unsolicited notifications about RSSI
                # changes, we will have to monitor it ourselves every 15s
                freq = 15

                daemon = SignalQualityDaemon(freq, device)
                col.append_daemon('signal', daemon)

            if N.SIG_NEW_CONN_MODE not in device.custom.device_capabilities:
                daemon = CellTypeDaemon(60, device)
                col.append_daemon('conn_mode', daemon)
        else:
            # device with just one port will never be able to send us
            # unsolicited notifications, we'll have to fake 'em
            daemon = SignalQualityDaemon(15, device)
            col.append_daemon('signal', daemon)

            daemon = CellTypeDaemon(60, device)
            col.append_daemon('conn_mode', daemon)

        return col
