# -*- coding: utf-8 -*-
# Copyright (C) 2006-2008  Vodafone España, S.A.
# Copyright (C) 2008  Warp Networks, S.L.
# Author:  Pablo Martí
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Dialer module abstracts the differences between dialers on different OSes
"""

import dbus
from dbus.service import Object, BusName, method

from zope.interface import implements

from twisted.internet import defer, reactor
from twisted.python import log

from wader.common.interfaces import IDialer
from wader.common.oal import osobj 
from wader.common.hardware import DBusExporterHelper
from wader.common.configbase import WaderConfigBase
import wader.common.consts as consts
from wader.utils.utilities import save_file

class DialerConf(object):
    """
    I contain all the necessary information to connect to Internet
    """
    apn = None
    username = None
    password = None
    connection = None
    staticdns = False
    dns1 = None
    dns2 = None

    def __init__(self):
        super(DialerConf, self).__init__()

    def __repr__(self):
        msg = '<DialerConf instance apn: %s, user: %s, passwd: %s>'
        args = (self.apn, self.username, self.password)
        return msg % args

    def __str__(self):
        return self.__repr__()

    @classmethod
    def from_dbus_path(cls, path):
        """
        Returns a new DialerConf instance from the C{path} opath
        """
        import os
        _id = path.split('/')[-1]
        prof_path = os.path.join(consts.MOBILE_PROFILES, _id)

        prof = WaderConfigBase(prof_path)

        for k, v in prof.conf.items('connection'):
            setattr(cls, k, v)

        return cls


class Dialer:
    """
    Base Dialer class

    Override me for new OSes
    """
    implements(IDialer)
    config = None
    protocol = None

    def configure(self, config, device):
        """
        Configures C{device} with C{config}

        This method should perform any necessary actions to connect to
        Internet like generating configuration files, modifying any necessary
        files, etc.
        """
        raise NotImplementedError()

    def connect(self):
        """
        Connects to Internet

        Returns a C{Deferred} that will be callbacked when connected

        @raise wader.common.exceptions.AlreadyConnected:
        @raise wader.common.exceptions.AlreadyConnecting:
        """
        raise NotImplementedError()

    def disconnect(self):
        """
        Disconnects from Internet

        Returns a C{Deferred} that will be callbacked when disconnected
        """
        raise NotImplementedError()


class DialerManager(Object, DBusExporterHelper):

    def __init__(self, ctrl):
        self.bus = dbus.SessionBus()
        name = BusName(consts.WADER_DIALUP_SERVICE, bus=self.bus)
        super(DialerManager, self).__init__(bus_name=name,
                                    object_path=consts.WADER_DIALUP_OBJECT)
        self.dialers = {}
        self.ctrl = ctrl
        self.connect_to_signals()

    def device_removed_cb(self, udi):
        if udi in self.dialers:
            log.msg("Device %s removed! deleting dialer instance" % udi)
            try:
                self.deactivate_connection(udi)
            except RuntimeError:
                pass

    def connect_to_signals(self):
        self.bus.add_signal_receiver(self.device_removed_cb,
                                     "DeviceRemoved",
                                     consts.WADER_INTFACE)

    def get_dialer(self, dev_opath):
        """
        Returns an instance of the dialer that must be used for this device/distro
        """
        from wader.common.dialers.wvdial import WVDialDialer
        from wader.common.dialers.hsolink import HSODialer
        plugin = self.ctrl.clients[dev_opath]
        # as dev_opath is already exported, we need a different objpath
        dial_path = "%s/Dial" % dev_opath

        if plugin.dialer == 'hsolinkcontrol':
            return HSODialer(dial_path, self.ctrl)
        else:
            return WVDialDialer(dial_path, self.ctrl)

    def configure_device(self, profile_path, device_path):
        from wader.common.hardware import CONN_OPTS_REV
        conf = DialerConf.from_dbus_path(profile_path)
        plugin = self.ctrl.clients[device_path]

        opts = dict(conn=CONN_OPTS_REV[conf.connection])
        return plugin.sconn.configure_connection(opts)

    def activate_connection(self, profile_path, device_path):
        dialer = self.get_dialer(device_path)

        deferred = defer.Deferred()

        def after_configuring_device_connect():
            self.dialers[device_path] = dialer
            d = dialer.configure(profile_path)
            d.addCallback(lambda ign: dialer.connect())
            d.chainDeferred(deferred)

        d = self.configure_device(profile_path, device_path)
        d.addCallback(lambda ign:
                    reactor.callLater(3, after_configuring_device_connect))
        return deferred

    def deactivate_connection(self, device_path):
        if not device_path in self.dialers:
            raise RuntimeError("Dialup %s not handled" % device_path)

        dialer = self.dialers[device_path]
        return dialer.disconnect()

    def get_stats_for_dialer(self, device_path):
        dialer = self.dialers[device_path]
        stats = osobj.get_iface_stats(dialer.iface)
        return stats

    @method(consts.DIL_INTFACE, in_signature='oo', out_signature='o',
                        async_callbacks=('async_cb', 'async_eb'))
    def ActivateConnection(self, profile_path, device_path, async_cb, async_eb):
        d = self.activate_connection(profile_path, device_path)
        return self.add_generic_callbacks(d, async_cb, async_eb)

    @method(consts.DIL_INTFACE, in_signature='o', out_signature='',
                        async_callbacks=('async_cb', 'async_eb'))
    def DeactivateConnection(self, device_path, async_cb, async_eb):
        d = self.deactivate_connection(device_path)
        return self.add_generic_callbacks_and_swallow(d, async_cb, async_eb)

    @method(consts.DIL_INTFACE, in_signature='o', out_signature='(uu)')
    def GetStats(self, device_path):
        return self.get_stats_for_dialer(device_path)

def create_dns_lock(dns1, dns2):
    text = """DNS1 %s\nDNS2 %s\n""" % (dns1, dns2)
    save_file(consts.WADER_DNS_LOCK, text)

