# -*- coding: utf-8 -*-
# Copyright (C) 2008  Warp Networks, S.L.
# Author:  Pablo Martí
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
HSOLink Dialer
"""

import re

import dbus
from dbus.service import Object, BusName, signal

from twisted.python import log, procutils
from twisted.internet import reactor, defer, protocol

import wader.common.consts as consts
from wader.common.dialer import Dialer, DialerConf


class HSODialer(Object, Dialer):
    """
    Dialer for HSO devices
    """
    binary = 'hsolinkcontrol'

    def __init__(self, dev_opath, ctrl):
        name = BusName(consts.WADER_DIALUP_SERVICE, bus=dbus.SessionBus())
        super(HSODialer, self).__init__(bus_name=name,
                                        object_path=dev_opath)
        self.iface = 'hso0'
        self.dev_opath = dev_opath
        self.bin_path = procutils.which(self.binary)[0]
        self.conf_path = ""
        self.num_of_retries = 0
        self.ctrl = ctrl
        self.proto = None
        self.iconn = None
        self.dns1 = None
        self.dns2 = None
        self.ip = None

    @signal(dbus_interface=consts.DIL_INTFACE, signature='')
    def Connected(self):
        log.msg("Emitting Connected signal")

    @signal(dbus_interface=consts.DIL_INTFACE, signature='')
    def Disconnected(self):
        log.msg("Emitting Disconnected signal")

    @signal(dbus_interface=consts.DIL_INTFACE, signature='as')
    def InvalidDNS(self, dns):
        log.msg("Emitting InvalidDNS signal %s" % dns)

    def check_assumptions(self):
        return []

    def check_permissions(self):
        return []

    def configure(self, prof_opath):
        return self._generate_config(prof_opath, self.dev_opath)

    def last_conf(self, ignored):
        self.device.sconn.send_at('AT_OHCIP')
        return self.dev_opath

    def connect(self):
        self.proto = HSOLinkProtocol(self)
        args = [self.binary, 'up', '0', self.dns1, self.dns2, self.ip]
        self.iconn = reactor.spawnProcess(self.proto, args[0], args, env=None)
        d = self.proto.deferred
        d.addCallback(self.last_conf)
        return d

    def disconnect(self):

        def kill_it():
            import os
            args = [self.binary, 'down', '0']
            os.system("%s %s %s" % tuple(args))

        def disconnect_cb(ignored=None):
            self._cleanup()
            return self.device.sconn.send_at('AT_OWANCALL=1,0,1')

        from twisted.internet.threads import deferToThread

        d = deferToThread(kill_it)
        d.addCallback(disconnect_cb)
        return d

    def _generate_config(self, conf_opath, device_opath):
        # as this is living in the same process, we cannot rely on DBus
        # for it, and we gotta use python directly
        conf = DialerConf.from_dbus_path(conf_opath)

        apn = conf.apn
        user = conf.username
        passwd = conf.password

        d = defer.Deferred()

        self.device = self.ctrl.clients[device_opath.replace('/Dial', '')]
        sconn = self.device.sconn
        sconn.send_at('AT+CGDCONT=1,"IP", "%s"' % apn)
        sconn.send_at('AT$QCPDPP=1,1,"%s","%s"' % (user, passwd))
        sconn.send_at('AT_OWANCALL=1,1,1')

        def get_config():
            def get_config_cb(confstr):
                confstr = confstr.replace(' ', '')
                data = confstr.split(',')

                self.ip, dns1, dns2 = data[1], data[3], data[4]

                if conf.staticdns:
                    self.dns1, self.dns2 = conf.dns1, conf.dns2
                else:
                    self.dns1, self.dns2 = dns1, dns2

                d.callback(self.dev_opath)

            def get_config_eb(error):
                log.err(error)
                self.num_of_retries += 1
                if self.num_of_retries < 3:
                    reactor.callLater(3, get_config)
                else:
                    self._cleanup()
                    d.errback(error)

            d2 = sconn.send_at('AT_OWANDATA=1')
            d2.addCallback(get_config_cb)
            d2.addErrback(get_config_eb)

        # give get_config 2 seconds, otherwise will fail
        reactor.callLater(3, get_config)
        return d

    def _cleanup(self):
        """cleanup our traces"""
        try:
            self.remove_from_connection()
        except:
            pass


ADDED_ROUTE = re.compile('added route')

class HSOLinkProtocol(protocol.ProcessProtocol):
    """ProcessProtocol for hsolinkcontrol"""

    def __init__(self, dialer):
        self.dialer = dialer
        self.__connected = False
        self.deferred = defer.Deferred()
        self.dns = []

    def connectionMade(self):
        self.transport.closeStdin()

    def outReceived(self, data):
        log.msg("HSOLink: sysout %s" % data)
        self.parse_output(data)

    def errReceived(self, data):
        log.msg('HSOLink: DATA RECV %s' % data)

    def outConnectionLost(self):
        log.msg('HSOLink: closed their stdout!')

    def errConnectionLost(self):
        log.msg('HSOLink: closed their stderr.')

    def processEnded(self, status_object):
        log.msg('HSOLink: quitting')

    def parse_output(self, data):
        m = ADDED_ROUTE.search(data)
        if m:
            log.msg("Emitting Connected signal")
            self.deferred.callback(True)
            self.dialer.Connected()

