# -*- coding: utf-8 -*-
# Copyright (C) 2006-2008  Vodafone España, S.A.
# Copyright (C) 2008  Warp Networks, S.L.
# Author:  Pablo Martí
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
WVdial Dialer
"""

from cStringIO import StringIO
import os
import re
from string import Template
import tempfile

import dbus
from dbus.service import Object, BusName, signal

from twisted.python import log, procutils
from twisted.internet import utils, reactor, defer, protocol, error

import wader.common.consts as consts
from wader.common.dialer import Dialer, DialerConf, create_dns_lock
from wader.common.hardware import DBusComponent
from wader.utils.utilities import get_file_data, save_file, is_bogus_ip

PAP_SECRETS = os.path.join(consts.TOP_DIR, 'etc', 'ppp', 'pap-secrets')
CHAP_SECRETS = os.path.join(consts.TOP_DIR, 'etc', 'ppp', 'chap-secrets')

DEFAULT_TEMPLATE = """
debug
noauth
name wvdial
replacedefaultroute
noipdefault
nomagic
usepeerdns
ipcp-accept-local
ipcp-accept-remote
nomp
noccp
nopredictor1
novj
novjccomp
nobsdcomp"""

PAP_TEMPLATE = DEFAULT_TEMPLATE + """
refuse-chap
refuse-mschap
refuse-mschap-v2
refuse-eap
"""

CHAP_TEMPLATE = DEFAULT_TEMPLATE + """
refuse-pap
"""

TEMPLATES_DICT = {
    'default' : DEFAULT_TEMPLATE,
    'PAP' : PAP_TEMPLATE,
    'CHAP' : CHAP_TEMPLATE,
}

### wvdial.conf stuff
def get_wvdial_conf_file(conf, serial_port):
    """Returns the path of the generated wvdial.conf"""
    text = _generate_wvdial_conf(conf, serial_port)
    dirpath = tempfile.mkdtemp('', 'Wader', '/tmp')
    filepath = tempfile.mkstemp('wvdial.conf', 'Wader', dirpath, True)[1]
    save_file(filepath, text)
    return filepath

def _generate_wvdial_conf(conf, sport):
    """
    Generates a specially crafted wvdial.conf with C{conf} and C{sport}
    """
    user = conf.username
    passwd = conf.password
    theapn = conf.apn

    # build template
    data = StringIO(get_file_data(consts.WVTEMPLATE))
    template = Template(data.getvalue())
    data.close()
    # return template
    opts = dict(serialport=sport, username=user, password=passwd, apn=theapn)
    return template.substitute(opts)


class WVDialDialer(Object, Dialer):
    """
    Dialer for Wvdial
    """
    binary = 'wvdial'

    def __init__(self, dev_opath, ctrl):
        name = BusName(consts.WADER_DIALUP_SERVICE, bus=dbus.SessionBus())
        super(WVDialDialer, self).__init__(bus_name=name,
                                           object_path=dev_opath)
        self.dev_opath = dev_opath
        self.bin_path = procutils.which(self.binary)[0]
        self.conf_path = ""
        self.ctrl = ctrl
        self.proto = None
        self.iconn = None
        self.iface = 'ppp0'

    @signal(dbus_interface=consts.DIL_INTFACE, signature='')
    def Connected(self):
        log.msg("Emitting Connected signal")

    @signal(dbus_interface=consts.DIL_INTFACE, signature='')
    def Disconnected(self):
        log.msg("Emitting Disconnected signal")

    @signal(dbus_interface=consts.DIL_INTFACE, signature='as')
    def InvalidDNS(self, dns):
        log.msg("Emitting InvalidDNS signal %s" % dns)

    def check_assumptions(self):
        return []

    def check_permissions(self):
        return []

    def configure(self, prof_opath):
        return defer.maybeDeferred(self._generate_config,
                                   prof_opath, self.dev_opath)

    def connect(self):
        self.proto = WVDialProtocol(self)
        args = [self.binary, '-C', self.conf_path, 'connect']
        self.iconn = reactor.spawnProcess(self.proto, args[0], args, env=None)
        return self.proto.deferred

    def disconnect(self):
        # ignore the fact that we are gonna be disconnected
        if not self.proto:
            return defer.succeed(self.dev_opath)

        self.proto.ignore_disconnect = True

        try:
            self.proto.transport.signalProcess('KILL')
        except error.ProcessExitedAlready:
            return defer.succeed(self.dev_opath)

        # just be damn sure that we're killing everything
        if self.proto.pid:
            kill_path = procutils.which('kill')[0]
            args = [kill_path, '-9', self.proto.pid]

        d = utils.getProcessValue(args[0], args, env=None)
        def disconnect_cb(error_code):
            log.err("EXIT CODE %d" % error_code)
            self._cleanup()
            return defer.succeed(self.dev_opath)

        d.addCallback(disconnect_cb)
        d.addErrback(lambda failure: log.err(failure))
        return d

    def _generate_config(self, conf_opath, device_opath):
        # generate wvdial.conf from template
        co = DBusComponent()
        device_opath = device_opath.replace('/Dial', '')
        device = self.ctrl.clients[device_opath]

        dport = device.dport

        # as this is living in the same process, we cannot rely on DBus
        # for it, and we gotta use python directly
        conf = DialerConf.from_dbus_path(conf_opath)
        self.conf_path = get_wvdial_conf_file(conf, dport)

        if conf.staticdns == '1':
            dns1 = conf.dns1
            dns2 = conf.dns2

            create_dns_lock(dns1, dns2)

    def _cleanup(self):
        """cleanup our traces"""
        try:
            path = os.path.dirname(self.conf_path)
            os.unlink(self.conf_path)
            os.rmdir(path)
        except (IOError, OSError):
            pass

        self.remove_from_connection()


PPPD_PID_REGEXP = re.compile('Pid of pppd: (?P<pid>\d+)')
PPPD_IFACE_REGEXP = re.compile('Using interface (?P<iface>ppp\d+)')
MAX_ATTEMPTS_REGEXP = re.compile('Maximum Attempts Exceeded')
PPPD_DIED_REGEXP = re.compile('The PPP daemon has died')
DNS_REGEXP = re.compile(r"""
   DNS\saddress
   \s                                     # beginning of the string
   (?P<ip>                                # group named ip
   (25[0-5]|                              # integer range 250-255 OR
   2[0-4][0-9]|                           # integer range 200-249 OR
   [01]?[0-9][0-9]?)                      # any number < 200
   \.                                     # matches '.'
   (25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?) # repeat
   \.                                     # matches '.'
   (25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?) # repeat
   \.                                     # matches '.'
   (25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?) # repeat
   )                                      # end of group
   \b                                     # end of the string
   """, re.VERBOSE)

class WVDialProtocol(protocol.ProcessProtocol):
    """ProcessProtocol for wvdial"""

    def __init__(self, dialer):
        self.dialer = dialer
        self.__connected = False
        self.pid = None
        self.number_of_inits = 0
        self.deferred = defer.Deferred()
        self.ignore_disconnect = False
        self.dns = []

    def connectionMade(self):
        self.transport.closeStdin()

    def outReceived(self, data):
        log.msg("WVDIAL: sysout %s" % data)

    def errReceived(self, data):
        """wvdial has this bad habit of using stderr for debug"""
        log.msg('WVDIAL: DATA RECV %s' % data)
        self.parse_output(data)

    def outConnectionLost(self):
        log.msg('WVDIAL: pppd closed their stdout!')

    def errConnectionLost(self):
        log.msg('WVDIAL: pppd closed their stderr.')

    def processEnded(self, status_object):
        log.msg('WVDIAL: quitting')
        if not self.__connected:
            if not self.ignore_disconnect:
                self.dialer.disconnect()
                self.dialer.Disconnected()

    def extract_iface(self, data):
        m = PPPD_IFACE_REGEXP.search(data)
        if m:
            self.dialer.iface = m.group('iface')
            log.msg("WVDIAL: dialer interface %s" % self.dialer.iface)

    def extract_dns_strings(self, data):
        if self.__connected:
            return

        for match in re.finditer(DNS_REGEXP, data):
            dns_ip = match.group('ip')
            self.dns.append(dns_ip)

        if len(self.dns) == 2:
            # Notify the user we are connected
            self.dialer.Connected()
            self.deferred.callback(self.dialer.dev_opath)
            self.__connected = True
            # check if they're valid DNS ips
            if is_bogus_ip(self.dns[0]) or is_bogus_ip(self.dns[1]):
                # the DNS assigned by the APN is probably invalid
                # lets notify the user only if she didn't specify static DNS
                self.dialer.InvalidDNS(self.dns)

    def extract_disconnected(self, data):
        # more than three attempts
        disconnected = MAX_ATTEMPTS_REGEXP.search(data)
        # pppd died
        pppd_died = PPPD_DIED_REGEXP.search(data)
        # extrac pppd pid
        pid = PPPD_PID_REGEXP.search(data)
        if pid:
            self.pid = pid.group('pid')
            self.number_of_inits += 1
        # wvdial refuses to stop after three attempts?
        if disconnected or pppd_died or self.number_of_inits >= 3:
            if not self.ignore_disconnect:
                self.dialer.disconnect()
                self.dialer.Disconnected()

    def parse_output(self, data):
        self.extract_iface(data)
        self.extract_dns_strings(data)
        self.extract_disconnected(data)

