# -*- coding: utf-8 -*-
# Copyright (C) 2008 Warp Networks S.L.
# Author:  Pablo Martí
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Wrappers around L{CDMAWrapper} and L{WCDMAWrapper}

I export their methods over DBus
"""

from dbus.service import method, signal

from twisted.python import log

from wader.common.consts import (SMS_INTFACE, CTS_INTFACE, PIN_INTFACE,
                               NET_INTFACE, CRD_INTFACE)
from wader.common.middleware import CDMAWrapper, WCDMAWrapper
from wader.common.sms import Message
from wader.common.persistent import Contact

# welcome to the multiple inheritance madness!
# python-dbus currently lacks an "export_as" keyword for use cases like
# us. Where we have a main object with dozens of methods that we want to
# export over several interfaces under repeated names, such as:
#   es.warp.Wader.Contacts.List
#   es.warp.Wader.SMS.List

# currently python-dbus requires you to create a new class and it will find
# the appropiated implementation through the MRO. But this leads to MH madness


class BaseExporterMixin(object):
    """
    Common methods for both L{WCDMAAdapter} and L{CDMAAdapter}
    """
    # es.warp.Wader.Card methods

    @method(CRD_INTFACE, in_signature='a{su}', out_signature='',
            async_callbacks=('async_cb', 'async_eb'))
    def ConfigureConnection(self, opts, async_cb, async_eb):
        """
        Configures the connection with the given parameters
        """
        d = self.configure_connection(opts)
        return self.add_generic_callbacks_and_swallow(d, async_cb, async_eb)

    @method(CRD_INTFACE, in_signature='', out_signature='',
            async_callbacks=('async_cb', 'async_eb'))
    def DisableEcho(self, async_cb, async_eb):
        """
        Disables echo
        """
        d = self.disable_echo()
        return self.add_generic_callbacks_and_swallow(d, async_cb, async_eb)

    @method(CRD_INTFACE, in_signature='', out_signature='',
            async_callbacks=('async_cb', 'async_eb'))
    def EnableEcho(self, async_cb, async_eb):
        """
        Enables echo

        Enabling echo will leave your connection unusable as this
        application assumes that it will be disabled
        """
        d = self.enable_echo()
        return self.add_generic_callbacks_and_swallow(d, async_cb, async_eb)

    @method(CRD_INTFACE, in_signature='', out_signature='',
            async_callbacks=('async_cb', 'async_eb'))
    def Enable(self, async_cb, async_eb):
        """
        Performs some initial setup in the device

        This method should always be called right after an AddDevice
        otherwise commands on the device might horribly fail
        """
        d = self._enable_device()
        return self.add_generic_callbacks_and_swallow(d, async_cb, async_eb)

    @method(CRD_INTFACE, in_signature='', out_signature='au',
            async_callbacks=('async_cb', 'async_eb'))
    def GetBands(self, async_cb, async_eb):
        """
        Returns a list with the available bands for the device
        """
        d = self.get_bands()
        return self.add_generic_callbacks(d, async_cb, async_eb)

    @method(CRD_INTFACE, in_signature='', out_signature='s',
            async_callbacks=('async_cb', 'async_eb'))
    def GetManufacturer(self, async_cb, async_eb):
        """
        Returns the device manufacturer name
        """
        d = self.get_manufacturer_name()
        return self.add_generic_callbacks(d, async_cb, async_eb)

    @method(CRD_INTFACE, in_signature='', out_signature='s',
            async_callbacks=('async_cb', 'async_eb'))
    def GetModel(self, async_cb, async_eb):
        """
        Returns the device model
        """
        d = self.get_card_model()
        return self.add_generic_callbacks(d, async_cb, async_eb)

    @method(CRD_INTFACE, in_signature='', out_signature='s',
            async_callbacks=('async_cb', 'async_eb'))
    def GetVersion(self, async_cb, async_eb):
        """
        Returns the device firmware version
        """
        d = self.get_card_version()
        return self.add_generic_callbacks(d, async_cb, async_eb)

    @method(CRD_INTFACE, in_signature='s', out_signature='s',
            async_callbacks=('async_cb', 'async_eb'))
    def SendATString(self, at_str, async_cb, async_eb):
        """
        Sends C{at_str} to the SIM
        """
        d = self.send_at(at_str)
        return self.add_generic_callbacks(d, async_cb, async_eb)

    @method(CRD_INTFACE, in_signature='', out_signature='',
            async_callbacks=('async_cb', 'async_eb'))
    def ResetSettings(self, async_cb, async_eb):
        """
        Resets the stored settings in SIM
        """
        d = self.reset_settings()
        return self.add_generic_callbacks_and_swallow(d, async_cb, async_eb)

    # es.warp.Wader.Network methods

    @method(NET_INTFACE, in_signature='', out_signature='i',
            async_callbacks=('async_cb', 'async_eb'))
    def GetSignal(self, async_cb, async_eb):
        """
        Returns the signal quality
        """
        d = self.get_signal_level()
        return self.add_generic_callbacks(d, async_cb, async_eb)

    # Signals

    @signal(dbus_interface=NET_INTFACE, signature='s')
    def CellTypeChanged(self, newcell):
        log.msg("%s: Emitting CellTypeChanged %s" % (self, newcell))

    @signal(dbus_interface=NET_INTFACE, signature='u')
    def CregReceived(self, status):
        log.msg("%s: Emitting CregReceived %d" % (self, status))

    @signal(dbus_interface=NET_INTFACE, signature='u')
    def RSSIChanged(self, rssi):
        log.msg("%s: Emitting RSSIChanged %d" % (self, rssi))

    @signal(dbus_interface=NET_INTFACE, signature='(ss)')
    def SpeedChanged(self, (up, down)):
        log.msg("%s: Emitting SpeedChanged (%s, %s)" % (self, up, down))

    @signal(dbus_interface=SMS_INTFACE, signature='us')
    def SMSReceived(self, index, where):
        log.msg('%s: Emitting SMSReceived %d at "%s"' % (self, index, where))

    @signal(dbus_interface=CRD_INTFACE, signature='o')
    def DeviceEnabled(self, opath):
        log.msg('%s: Emitting DeviceEnabled "%s"' % (self, opath))


class ContactsExporter(WCDMAWrapper):
    """
    es.warp.Wader.Contacts methods
    """

    @method(CTS_INTFACE, in_signature='ss', out_signature='u',
            async_callbacks=('async_cb', 'async_eb'))
    def Add(self, name, number, async_cb, async_eb):
        """
        Adds a contact with C{name} and C{number} and returns the index
        """
        d = self.add_contact(Contact(name, number))
        return self.add_generic_callbacks(d, async_cb, async_eb)

    @method(CTS_INTFACE, in_signature='u', out_signature='',
            async_callbacks=('async_cb', 'async_eb'))
    def Delete(self, index, async_cb, async_eb):
        """
        Deletes the contact at C{index}
        """
        d = self.delete_contact(index)
        return self.add_generic_callbacks_and_swallow(d, async_cb, async_eb)

    @method(CTS_INTFACE, in_signature='u', out_signature='(uss)',
            async_callbacks=('async_cb', 'async_eb'))
    def GetByIndex(self, index, async_cb, async_eb):
        """
        Returns the contact at C{index}

        Returns a struct with index, name and number of the contact
        """
        d = self.get_contact_by_index(index)
        d.addCallback(lambda c: (c.index, c.name, c.number))
        return self.add_generic_callbacks(d, async_cb, async_eb)

    @method(CTS_INTFACE, in_signature='', out_signature='a(uss)',
            async_callbacks=('async_cb', 'async_eb'))
    def List(self, async_cb, async_eb):
        """
        Returns all the contacts in the SIM

        Returns an array of structs with index, name and number
        """
        d = self.get_contacts()
        d.addCallback(lambda contacts:
                      [(c.index, c.name, c.number) for c in contacts])
        return self.add_generic_callbacks(d, async_cb, async_eb)


class PINAndSMSExporter(ContactsExporter):
    """
    es.warp.Wader.{PIN,SMS} methods and signals
    """

    @method(SMS_INTFACE, in_signature='u', out_signature='',
            async_callbacks=('async_cb', 'async_eb'))
    def Delete(self, index, async_cb, async_eb):
        """
        Deletes the SMS at C{index}
        """
        d = self.delete_sms(index)
        return self.add_generic_callbacks_and_swallow(d, async_cb, async_eb)

    @method(SMS_INTFACE, in_signature='', out_signature='a(uss)',
            async_callbacks=('async_cb', 'async_eb'))
    def List(self, async_cb, async_eb):
        """
        Returns all the SMS stored in SIM
        """
        d = self.get_sms()
        d.addCallback(lambda messages:
                    [(sms.index, sms.text, sms.number) for sms in messages])
        return self.add_generic_callbacks(d, async_cb, async_eb)

    @method(SMS_INTFACE, in_signature='u', out_signature='(uss)',
            async_callbacks=('async_cb', 'async_eb'))
    def GetByIndex(self, index, async_cb, async_eb):
        """
        Returns the SMS stored at C{index}
        """
        d = self.get_sms_by_index(index)
        d.addCallback(lambda sms: (sms.index, sms.text, sms.number))
        return self.add_generic_callbacks(d, async_cb, async_eb)

    @method(SMS_INTFACE, in_signature='ss', out_signature='u',
            async_callbacks=('async_cb', 'async_eb'))
    def Save(self, number, text, async_cb, async_eb):
        """
        Save a SMS with C{text} and C{number} and returns the index
        """
        d = self.save_sms(Message(number, text))
        return self.add_generic_callbacks(d, async_cb, async_eb)

    @method(SMS_INTFACE, in_signature='ss', out_signature='u',
            async_callbacks=('async_cb', 'async_eb'))
    def Send(self, number, text, async_cb, async_eb):
        """
        Sends a SMS with C{text} to C{number}
        """
        d = self.send_sms(Message(number, text))
        return self.add_generic_callbacks(d, async_cb, async_eb)

    @method(SMS_INTFACE, in_signature='u', out_signature='u',
            async_callbacks=('async_cb', 'async_eb'))
    def SendFromStorage(self, index, async_cb, async_eb):
        """
        Sends the SMS stored at C{index} and returns the new index
        """
        d = self.send_sms_from_storage(index)
        return self.add_generic_callbacks(d, async_cb, async_eb)


    # es.warp.Wader.PIN methods

    @method(PIN_INTFACE, in_signature='s', out_signature='',
            async_callbacks=('async_cb', 'async_eb'))
    def Enable(self, pin, async_cb, async_eb):
        """
        Enables PIN authentication using C{pin}
        """
        d = self.enable_pin(pin)
        return self.add_generic_callbacks_and_swallow(d, async_cb, async_eb)


class WCDMAAdapter(BaseExporterMixin, PINAndSMSExporter):
    """
    I export L{WCDMAWrapper} methods over DBus
    """

    def __str__(self):
        """
        We override __str__ coz dbus hijacks __repr__ and the log gets messy
        """
        return "WCDMAAdapter"

    __repr__ = __str__

    # es.warp.Wader.Card methods

    @method(CRD_INTFACE, in_signature='', out_signature='as',
            async_callbacks=('async_cb', 'async_eb'))
    def GetAvailableCharsets(self, async_cb, async_eb):
        """
        Returns the available charsets in SIM
        """
        d = self.get_available_charset()
        return self.add_generic_callbacks(d, async_cb, async_eb)

    @method(CRD_INTFACE, in_signature='', out_signature='s',
            async_callbacks=('async_cb', 'async_eb'))
    def GetCharset(self, async_cb, async_eb):
        """
        Returns active charset
        """
        d = self.get_charset()
        return self.add_generic_callbacks(d, async_cb, async_eb)

    @method(CRD_INTFACE, in_signature='', out_signature='s',
            async_callbacks=('async_cb', 'async_eb'))
    def GetIMEI(self, async_cb, async_eb):
        """
        Returns the device's IMEI
        """
        d = self.get_imei()
        return self.add_generic_callbacks(d, async_cb, async_eb)

    @method(CRD_INTFACE, in_signature='', out_signature='s',
            async_callbacks=('async_cb', 'async_eb'))
    def GetIMSI(self, async_cb, async_eb):
        """
        Returns the SIM's IMSI
        """
        d = self.get_imsi()
        return self.add_generic_callbacks(d, async_cb, async_eb)

    @method(CRD_INTFACE, in_signature='', out_signature='i',
            async_callbacks=('async_cb', 'async_eb'))
    def GetPhonebookSize(self, async_cb, async_eb):
        """
        Returns the phonebook size
        """
        d = self.get_phonebook_size()
        return self.add_generic_callbacks(d, async_cb, async_eb)

    @method(CRD_INTFACE, in_signature='s', out_signature='',
            async_callbacks=('async_cb', 'async_eb'))
    def SetCharset(self, charset, async_cb, async_eb):
        """
        Sets the SIM charset to C{charset}
        """
        d = self.set_charset(charset.encode('utf8'))
        return self.add_generic_callbacks_and_swallow(d, async_cb, async_eb)

    # es.warp.Wader.Network methods

    @method(NET_INTFACE, in_signature='', out_signature='(uu)',
            async_callbacks=('async_cb', 'async_eb'))
    def GetRegistrationStatus(self, async_cb, async_eb):
        """
        Returns the network registration status
        """
        d = self.get_netreg_status()
        return self.add_generic_callbacks(d, async_cb, async_eb)

    @method(NET_INTFACE, in_signature='', out_signature='(ss)',
            async_callbacks=('async_cb', 'async_eb'))
    def GetInfo(self, async_cb, async_eb):
        """
        Returns the network information

        Returns a struct with CurrentOperatorName, link type
        """
        d = self.get_network_info()
        return self.add_generic_callbacks(d, async_cb, async_eb)

    @method(NET_INTFACE, in_signature='', out_signature='a(ussuu)',
            async_callbacks=('async_cb', 'async_eb'))
    def GetNames(self, async_cb, async_eb):
        """
        Returns the network information (aka AT+COPS=?)
        """
        d = self.get_network_names()
        d.addCallback(lambda netobjs:
                [(n.stat, n.short_name, n.long_name, n.netid, n.rat)
                                        for n in netobjs])
        return self.add_generic_callbacks(d, async_cb, async_eb)

    @method(NET_INTFACE, in_signature='', out_signature='as',
            async_callbacks=('async_cb', 'async_eb'))
    def GetRoamingIDs(self, async_cb, async_eb):
        """
        Returns an array of all the roaming IDs stored in the SIM
        """
        d = self.get_roaming_ids()
        d.addCallback(lambda objs: [obj.netid for obj in objs])
        return self.add_generic_callbacks(d, async_cb, async_eb)

    @method(NET_INTFACE, in_signature='', out_signature='i',
            async_callbacks=('async_cb', 'async_eb'))
    def GetSignal(self, async_cb, async_eb):
        """
        Returns the signal quality
        """
        d = self.get_signal_level()
        return self.add_generic_callbacks(d, async_cb, async_eb)

    @method(NET_INTFACE, in_signature='b', out_signature='',
            async_callbacks=('async_cb', 'async_eb'))
    def SetRegistrationNotification(self, active, async_cb, async_eb):
        """
        Sets the network registration notifications
        """
        d = self.set_netreg_notification(int(active))
        return self.add_generic_callbacks_and_swallow(d, async_cb, async_eb)

    @method(NET_INTFACE, in_signature='uu', out_signature='',
            async_callbacks=('async_cb', 'async_eb'))
    def SetInfoFormat(self, mode, format, async_cb, async_eb):
        """
        Sets the network info format
        """
        d = self.set_network_info_format(mode, format)
        return self.add_generic_callbacks_and_swallow(d, async_cb, async_eb)

    @method(NET_INTFACE, in_signature='u', out_signature='',
            async_callbacks=('async_cb', 'async_eb'))
    def RegisterWithNetID(self, netid, async_cb, async_eb):
        """
        Attemtps to register with given C{netid}
        """
        d = self.register_with_network(netid)
        return self.add_generic_callbacks_and_swallow(d, async_cb, async_eb)

    # es.warp.Wader.SMS methods

    @method(SMS_INTFACE, in_signature='', out_signature='s',
            async_callbacks=('async_cb', 'async_eb'))
    def GetSMSC(self, async_cb, async_eb):
        """
        Returns the SMSC number stored in the SIM
        """
        d = self.get_smsc()
        return self.add_generic_callbacks(d, async_cb, async_eb)

    @method(SMS_INTFACE, in_signature='', out_signature='u',
            async_callbacks=('async_cb', 'async_eb'))
    def GetFormat(self, async_cb, async_eb):
        """
        Returns 1 if SMS format is text and 0 if SMS format is PDU
        """
        d = self.get_sms_format()
        return self.add_generic_callbacks(d, async_cb, async_eb)

    @method(SMS_INTFACE, in_signature='u', out_signature='',
            async_callbacks=('async_cb', 'async_eb'))
    def SetFormat(self, format, async_cb, async_eb):
        """
        Sets the SMS format
        """
        if format not in [0, 1]:
            async_eb(ValueError("Invalid SMS format %s" % repr(format)))

        d = self.set_sms_format(format)
        return self.add_generic_callbacks_and_swallow(d, async_cb, async_eb)

    @method(SMS_INTFACE, in_signature='uuuuu', out_signature='',
            async_callbacks=('async_cb', 'async_eb'))
    def SetIndication(self, mode, mt, bm, ds, bfr, async_cb, async_eb):
        """
        Sets the SMS indication
        """
        d = self.set_sms_indication(mode, mt, bm, ds, bfr)
        return self.add_generic_callbacks_and_swallow(d, async_cb, async_eb)

    @method(SMS_INTFACE, in_signature='s', out_signature='',
            async_callbacks=('async_cb', 'async_eb'))
    def SetSMSC(self, smsc, async_cb, async_eb):
        """
        Sets the SIM SMSC to C{smsc}
        """
        d = self.set_smsc(smsc)
        return self.add_generic_callbacks_and_swallow(d, async_cb, async_eb)

    # es.warp.Wader.Contacts methods

    @method(CTS_INTFACE, in_signature='s', out_signature='a(uss)',
            async_callbacks=('async_cb', 'async_eb'))
    def Find(self, pattern, async_cb, async_eb):
        """
        Returns list of contacts that match C{pattern}

        Returns an array of structs with index, name and number
        """
        d = self.find_contacts(pattern)
        d.addCallback(lambda contacts:
                      [(c.index, c.name, c.number) for c in contacts])
        return self.add_generic_callbacks(d, async_cb, async_eb)

    # es.warp.Wader.PIN methods

    @method(PIN_INTFACE, in_signature='ss', out_signature='',
            async_callbacks=('async_cb', 'async_eb'))
    def Change(self, oldpin, newpin, async_cb, async_eb):
        """
        Changes PIN from C{oldpin} to C{newpin}
        """
        d = self.change_pin(oldpin, newpin)
        return self.add_generic_callbacks_and_swallow(d, async_cb, async_eb)

    @method(PIN_INTFACE, in_signature='', out_signature='s',
            async_callbacks=('async_cb', 'async_eb'))
    def Check(self, async_cb, async_eb):
        """
        Returns the SIM's auth state
        """
        d = self.check_pin()
        return self.add_generic_callbacks(d, async_cb, async_eb)

    @method(PIN_INTFACE, in_signature='s', out_signature='',
            async_callbacks=('async_cb', 'async_eb'))
    def Disable(self, pin, async_cb, async_eb):
        """
        Disables PIN authentication using C{pin}
        """
        d = self.disable_pin(pin)
        return self.add_generic_callbacks_and_swallow(d, async_cb, async_eb)

    @method(PIN_INTFACE, in_signature='s', out_signature='s',
            async_callbacks=('async_cb', 'async_eb'))
    def Send(self, pin, async_cb, async_eb):
        """
        Sends C{pin} to authenticate with SIM
        """
        d = self.send_pin(pin)
        d.addCallback(self._check_initted_device)
        return self.add_generic_callbacks(d, async_cb, async_eb)

    @method(PIN_INTFACE, in_signature='ss', out_signature='',
            async_callbacks=('async_cb', 'async_eb'))
    def SendPUK(self, puk, pin, async_cb, async_eb):
        """
        Sends C{puk} and C{pin} to authenticate with SIM
        """
        d = self.send_puk(puk, pin)
        d.addCallback(self._check_initted_device)
        return self.add_generic_callbacks_and_swallow(d, async_cb, async_eb)


class CDMAAdapter(BaseExporterMixin, CDMAWrapper):
    """
    I export L{CDMAWrapper} methods over DBus
    """

    def __str__(self):
        return "CDMAAdapter"

    __repr__ = __str__

    # es.warp.Wader.Card methods

    @method(CRD_INTFACE, in_signature='', out_signature='s',
            async_callbacks=('async_cb', 'async_eb'))
    def GetESN(self, async_cb, async_eb):
        """
        Returns the device ESN
        """
        d = self.get_esn()
        return self.add_generic_callbacks(d, async_cb, async_eb)

