# -*- coding: utf-8 -*-
# Copyright (C) 2006-2008  Vodafone España, S.A.
# Copyright (C) 2008  Warp Networks, S.L.
# Author:  Pablo Martí
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
The hardware module manages device discovery via dbus/hal on Unix/Linux
"""

import serial

from twisted.python import log
from twisted.internet import defer

from wader.common.hardware._dbus import DBusComponent
from wader.common.plugin import PluginManager
from wader.utils.utilities import natsort

def extract_info(props):
    info = {}
    if 'usb.vendor_id' in props:
        info['usb_device.vendor_id'] = props['usb.vendor_id']
        info['usb_device.product_id'] = props['usb.product_id']
    elif 'usb_device.vendor_id' in props:
        info['usb_device.vendor_id'] = props['usb_device.vendor_id']
        info['usb_device.product_id'] = props['usb_device.product_id']
    elif 'pcmia.manf_id' in props:
        info['pcmcia.manf_id'] = props['pcmcia.manf_id']
        info['pcmcia.card_id'] = props['pcmcia.card_id']
    elif 'pci.vendor_id' in props:
        info['pci.vendor_id'] = props['pci.vendor_id']
        info['pci.product_id'] = props['pci.product_id']
    else:
        raise RuntimeError("Unknown bus for device %s" % props['info.udi'])

    return info


class HardwareManager(DBusComponent):
    """
    I find and configure devices

    I am resilient to ports assigned in unusual locations
    and devices sharing ids.
    """

    def __init__(self):
        super(HardwareManager, self).__init__()

    def get_devices(self):
        """
        Returns a list with all the devices present in the system

        List of deferreds of course
        """
        parent_udis = self._get_parent_udis()
        return self._get_devices_from_udis(parent_udis)

    def get_device_from_udi(self, udi):
        """
        Returns a device built out of the info extracted from C{udi}
        """
        info = self._get_info_from_udi(udi)
        ports = self._get_ports_from_udi(udi)
        plugin = PluginManager.get_plugin_by_vendor_product_id(*info.values())
        dport = cport = None
        if plugin and plugin.dodgy:
            try:
                dport = ports[plugin.dodgy_ports[0]]
                cport = ports[plugin.dodgy_ports[1]]
            except IndexError:
                #if we get an index error the ports won't be correctly set
                # and that will be handled on _get_device_from_info_and_ports
                pass
        else:
            dport, cport = self._probe_ports(ports)
        device = self._get_device_from_info_and_ports(info, udi, dport, cport)
        return device

    def _get_devices_from_udis(self, udis):
        """
        Returns a list of devices built out of the info extracted from C{udis}
        """
        from wader.common.hardware.base import identify_device
        unknown_devs = map(self.get_device_from_udi, udis)
        deferreds = map(identify_device, unknown_devs)
        return defer.gatherResults(deferreds)

    def _get_parent_udis(self):
        """
        Returns the root udi of all the devices with modem capabilities
        """
        return set(map(self._get_parent_udi,
                          self.manager.FindDeviceByCapability("modem")))

    def _get_parent_udi(self, udi):
        """
        Returns the absolute parent udi of C{udi}
        """
        ORIG = 'serial.originating_device'
        def get_parent(props):
            if ORIG in props:
                return props[ORIG]
            return props['info.parent']

        current_udi = udi
        # log.msg("get_parent_udi called with %s" % udi)
        while True:
            # log.msg("obtaining properties from udi %s" % udi)
            props = self.get_properties_from_udi(current_udi)
            try:
                info = extract_info(props)
                # log.msg("Properties I'm looking for %s" % info)
                break
            except RuntimeError:
                current_udi = get_parent(props)

        # now that we have an id to lookup for, lets repeat the process till we
        # get another RuntimeError
        def find_out_if_contained(_info, props):
            """
            Returns C{True} if C{_info} values are contained in C{props}

            As hal likes to swap between usb.vendor_id and usb_device.vendor_id
            I have got a special case where I will retry
            """
            def compare_dicts(d1, d2):
                for key in d1.keys():
                    try:
                        return d1[key] == d2[key]
                    except KeyError:
                        return False

            if compare_dicts(_info, props):
                # we got a straight map
                return True
            # hal likes to swap between usb_device.vendor_id and usb.vendor_id
            if 'usb_device.vendor_id' in _info:
                # our last chance, perhaps its swapped
                newinfo = {'usb.vendor_id' : _info['usb_device.vendor_id'],
                           'usb.product_id' : _info['usb_device.product_id']}
                return compare_dicts(newinfo, props)

            # the original compare_dicts failed, so return False
            return False

        last_udi = current_udi
        while True:
            # log.msg("obtaining properties from udi (PART 2) %s" % current_udi)
            props = self.get_properties_from_udi(current_udi)
            if not find_out_if_contained(info, props):
                break
            parent_udi = get_parent(props)
            last_udi = current_udi
            current_udi = parent_udi
            # log.msg("The last udi is now %s" % parent_udi)

        return last_udi

    def _get_info_from_udi(self, udi):
        # log.msg("Obtaining info from udi %s" % udi)
        return extract_info(self.get_properties_from_udi(udi))

    def _get_ports_from_udi(self, udi):
        """
        Returns all the ports that C{udi} has registered
        """
        device_props = self.get_devices_properties()
        dev_udis = sorted(device_props.keys(), key=len)
        childs = []
        for _udi in dev_udis:
            if _udi != udi: # not ourselves
                try:
                    par_udi = device_props[_udi]['info.parent']
                except KeyError:
                    continue
                if par_udi == udi or par_udi in childs:
                    childs.append(_udi)

        if not childs:
            raise RuntimeError("Couldn't find any child of device %s" % udi)

        dp = device_props
        ports = map(str, [dp[_udi]['serial.device']
                        for _udi in childs if 'serial.device' in dp[_udi]])
        natsort(ports)
        return ports

    def _get_device_from_info_and_ports(self, info, udi, dport, cport):
        """
        Returns a C{DevicePlugin} out of C{info} and C{dport} and {cport}
        """
        if not dport and not cport:
            # this shouldn't happen
            raise RuntimeError("No data port and no control port")

        plugin = PluginManager.get_plugin_by_vendor_product_id(*info.values())
        if plugin:
            plugin.cport, plugin.dport = cport, dport
            plugin.udi = udi
            return plugin

        raise RuntimeError("Couldn't find a plugin with info %s" % info)

    def _probe_port(self, port):
        """
        Returns C{True} if C{port} works, otherwise returns C{False}
        """
        try:
            ser = serial.Serial(port, timeout=1)
            ser.write('AT+GMR\r\n')
            if not ser.readlines():
                # Huawei E620 with driver option registers three serial
                # ports and the middle one wont raise any exception while
                # opening it even thou its a dummy port.
                return False

            return True
        except serial.SerialException:
            return False
        finally:
            try:
                ser.close()
            except:
                pass

    def _probe_ports(self, ports):
        """
        Returns a tuple of (data port, control port) out of C{ports}
        """
        dport = cport = None
        while ports:
            port = ports.pop(0)
            if self._probe_port(port):
                if not dport:
                    # data port tends to the be the first one
                    dport = port
                elif not cport:
                    # control port the next one
                    cport = port
                    break

        return dport, cport

