# -*- coding: utf-8 -*-
# Copyright (C) 2006-2008  Vodafone España, S.A.
# Copyright (C) 2008  Warp Networks, S.L.
# Author:  Pablo Martí
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Base classes for the hardware module
"""

import serial
from twisted.internet.threads import deferToThread

from wader.common.command import get_cmd_dict_copy
from wader.common.plugin import PluginManager
from wader.common.statem.auth import AuthStateMachine
from wader.common.statem.connection import ConnectStateMachine
from wader.common.statem.networkreg import NetworkRegStateMachine
import wader.common.exceptions as ex


class CDMACustomizer(object):
    """
    I contain all the custom classes and metadata that a CDMA device needs

    @cvar adapter: Adapter for the device
    @type adapter: L{CDMAAdapter} child.
    @cvar async_regexp: regexp to parse asynchronous notifications emited
    by the device.
    @cvar conn_dict: Dictionary with the AT strings necessary to change
    between the different connection modes
    @cvar cmd_dict: Dictionary with commands info
    @cvar device_capabilities: List with the unsolicited notifications that
    this device supports
    @cvar authklass: Class that will handle the authentication for this device
    @cvar connklass: Class that will handle the connection for this device
    @cvar netrklass: Class that will handle the network registration for this
    device
    """
    from wader.common.exported import CDMAAdapter
    adapter = CDMAAdapter
    async_regexp = None
    band_dict = {}
    conn_dict = {}
    cmd_dict = get_cmd_dict_copy()
    device_capabilities = []
    signal_translations = {}
    authklass = AuthStateMachine
    connklass = ConnectStateMachine
    netrklass = NetworkRegStateMachine


class WCDMACustomizer(object):
    """
    I contain all the custom classes and metadata that a WCDMA device needs

    @cvar adapter: Adapter for the device
    @type adapter: L{WCDMAAdapter} child.
    @cvar async_regexp: regexp to parse asynchronous notifications emited
    by the device.
    @cvar conn_dict: Dictionary with the AT strings necessary to change
    between the different connection modes
    @cvar cmd_dict: Dictionary with commands info
    @cvar device_capabilities: List with the unsolicited notifications that
    this device supports
    @cvar authklass: Class that will handle the authentication for this device
    @cvar connklass: Class that will handle the connection for this device
    @cvar netrklass: Class that will handle the network registration for this
    device
    """
    from wader.common.exported import WCDMAAdapter
    adapter = WCDMAAdapter
    async_regexp = None
    band_dict = {}
    conn_dict = {}
    cmd_dict = get_cmd_dict_copy()
    device_capabilities = []
    signal_translations = {}
    authklass = AuthStateMachine
    connklass = ConnectStateMachine
    netrklass = NetworkRegStateMachine


def _identify_device(port):
    """
    Returns the model of the device present at C{port}
    """
    # as the readlines method blocks, this is executed in a parallel thread
    # with deferToThread
    ser = serial.Serial(port, timeout=1)
    ser.write('AT+GMM\r\n')
    response = ser.readlines()
    ser.close()
    return response[1].replace('\r\n', '')

def identify_device(plugin):
    def identify_device_cb(model):
        # plugin to return
        _plugin = None

        if plugin.mapping:
            if model not in plugin.mapping:
                # we've received a plugin with multiple personalities
                # and the default its not present, we can assume that the
                # default plugin in the mapping will do
                pass
            else:
                # the model is present in the plugin's mapping
                _plugin = plugin.mapping[model]()

        # the plugin has no mapping, chances are that we already identified
        # it by its vendor & product id
        elif plugin.__remote_name__ != model:
            # so we basically have a device identified by vendor & product id
            # but we know nuthin of this model
            try:
                _plugin = PluginManager.get_plugin_by_remote_name(model)
            except ex.UnknownPluginNameError:
                plugin.name = model

        if _plugin:
            # we found another plugin during the process
            _plugin.patch(plugin)
            plugin.close()
            return _plugin
        else:
            # the original plugin, free allocated resources and
            # return the same plugin
            plugin.close()
            return plugin

    d = deferToThread(_identify_device, plugin.dport)
    d.addCallback(identify_device_cb)
    return d

