# -*- coding: utf-8 -*-
# Copyright (C) 2006-2008  Vodafone España, S.A.
# Copyright (C) 2008  Warp Networks, S.L.
# Author:  Pablo Martí
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Common stuff for all Option's datacards/devices
"""

import re

from wader.common.command import get_cmd_dict_copy, OK_REGEXP, ERROR_REGEXP
from wader.common.hardware.base import WCDMACustomizer
from wader.common.sim import SIMBaseClass
from wader.common.plugin import DBusDevicePlugin
import wader.common.notifications as N

OPTION_CONN_DICT = {
    'ANY'    : 'AT_OPSYS=0',
    '2GONLY' : 'AT_OPSYS=0,2',
    '3GONLY' : 'AT_OPSYS=1,2',
    '2GPREF' : 'AT_OPSYS=2,2',
    '3GPREF' : 'AT_OPSYS=3,2',
}

OPTION_BAND_DICT = {
    'ANY'   : 'ANY,1',   # any band
    'EGSM'  : 'EGSM,1',  # EGSM (900MHz)
    'DCS'   : 'DCS,1',   # DCS (1800MHz)
    'PCS'   : 'PCS,1',   # PCS (1900MHz)
    'G850'  : 'G850,1',  # GSM (850 MHz)
    'U2100' : 'U2100,1', # WCDMA 2100Mhz            (Class I)
    'U1700' : 'U1700,1', # WCDMA 3GPP UMTS1800      (Class III)
    'U17IV' : 'U17IV,1', # WCDMA 3GPP AWS 1700/2100 (Class IV)
    'U850'  : 'U850,1',  # WCDMA 3GPP UMTS850       (Class V)
    'U800'  : 'U800,1',  # WCDMA 3GPP UMTS800       (Class VI)
    'U900'  : 'U900,1',  # WCDMA 3GPP UMTS900       (Class VIII)
    'U17IX' : 'U17IX,1', # WCDMA 3GPP UMTS          (Class IX)
}


# Option devices like to append its serial number after the IMEI, ignore it
OPTION_CMD_DICT = get_cmd_dict_copy()
info = dict(echo=None,
            end=OK_REGEXP,
            error=ERROR_REGEXP,
            extract=re.compile("\r\n(?P<imei>\d+),\S+\r\n"))

OPTION_CMD_DICT['get_imei'] = info

class OptionSIMClass(SIMBaseClass):
    """
    Option SIM Class

    I just activate unsolicited notifications for ya
    """
    def __init__(self, sconn):
        super(OptionSIMClass, self).__init__(sconn)

    def initialize(self, set_encoding=True):
        d = super(OptionSIMClass, self).initialize(set_encoding=set_encoding)
        def init_callback(size):
            # setup asynchronous notifications
            self.sconn.send_at('AT_OSSYS=1') # cell change notification
            self.sconn.send_at('AT_OSQI=1') # signal quality notification
            return size

        d.addCallback(init_callback)
        return d


def new_conn_mode_cb(args):
    """
    Translates Option's unsolicited notifications to Wader's representation
    """
    ossysi_args_dict = {
        '0' : N.GPRS_SIGNAL,
        '2' : N.UMTS_SIGNAL,
        '3' : N.NO_SIGNAL,
    }
    return ossysi_args_dict[args]


class OptionWCDMACustomizer(WCDMACustomizer):
    """Customizer for Option's cards"""
    async_regexp = re.compile(r"""
                \r\n
                (?P<signal>_O[A-Z]{3,}):\s(?P<args>.*)
                \r\n""", re.VERBOSE)
    band_dict = OPTION_BAND_DICT
    conn_dict = OPTION_CONN_DICT
    cmd_dict = OPTION_CMD_DICT
    device_capabilities = [N.SIG_NEW_CONN_MODE, N.SIG_RSSI]
    signal_translations = {
        '_OSSYSI' : (N.SIG_NEW_CONN_MODE, new_conn_mode_cb),
        '_OSIGQ'  : (N.SIG_RSSI, lambda args: int(args.split(',')[0]))
    }


class OptionWCDMADevicePlugin(DBusDevicePlugin):
    """DBusDevicePlugin for Option"""
    simklass = OptionSIMClass
    custom = OptionWCDMACustomizer()

    def __init__(self):
        super(OptionWCDMADevicePlugin, self).__init__()


