# -*- coding: utf-8 -*-
# Copyright (C) 2006-2008  Vodafone España, S.A.
# Copyright (C) 2008  Warp Networks, S.L.
# Author:  Pablo Martí
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Wrappers around L{CDMADeviceConnection} and L{WCDMADeviceConnection}

It basically provides error control and more high-level operations. N-tier
folks can see this as a Business Logic class.
"""

from collections import deque

from twisted.python import log
from twisted.internet import defer, reactor

from wader.common.encoding import (from_ucs2, from_u, unpack_ucs2_bytes,
                                pack_ucs2_bytes, check_if_ucs2, _)
import wader.common.exceptions as ex
from wader.cli.collaborator import DBusCollaboratorFactory
from wader.common.hardware import (DBusExporterHelper, CONN_OPTS, BAND_OPTS,
                                    BAND_OPTS_REV)
from wader.common.persistent import Contact
from wader.common.protocol import (BaseDeviceConnection, CDMADeviceConnection,
                                 WCDMADeviceConnection)
from wader.common.sms import pdu_to_message, message_to_pdu

def process_contact_match(match):
    """I process a contact match and return a C{Contact} object out of it"""
    name = from_ucs2(match.group('name'))
    number = from_ucs2(match.group('number'))
    index = int(match.group('id'))
    return Contact(name, number, index=index)


class BaseWrapper(BaseDeviceConnection, DBusExporterHelper):

    def disable_echo(self):
        d = super(BaseWrapper, self).disable_echo()
        d.addCallback(lambda result: result[0].group('resp'))
        return d

    def enable_echo(self):
        """
        Enables echo
        """
        d = super(BaseWrapper, self).enable_echo()
        d.addCallback(lambda result: result[0].group('resp'))
        return d

    def get_card_model(self):
        """
        Returns the card model

        @rtype: C{defer.Deferred}
        """
        d = super(BaseWrapper, self).get_card_model()
        d.addCallback(lambda response: response[0].group('model'))
        return d

    def get_card_version(self):
        """
        Returns the firmware version

        @rtype: C{defer.Deferred}
        """
        d = super(BaseWrapper, self).get_card_version()
        d.addCallback(lambda response: response[0].group('version'))
        return d

    def get_manufacturer_name(self):
        """
        Returns the Manufacturer name

        @rtype: C{Deferred}
        """
        d = super(BaseWrapper, self).get_manufacturer_name()
        d.addCallback(lambda response: response[0].group('name'))
        return d

    def get_signal_level(self):
        """
        Returns the signal level

        @rtype: C{Deferred}
        """
        d = super(BaseWrapper, self).get_signal_level()
        d.addCallback(lambda response: int(response[0].group('rssi')))
        return d

    def send_at(self, atstr):
        """
        Sends an arbitrary AT string C{atstr}
        """
        d = super(BaseWrapper, self).send_at(atstr)
        d.addCallback(lambda response: response[0].group('resp'))
        return d

    # common configure_connection methods

    def get_bands(self):
        """
        Returns a list with the available bands
        """
        if not self.custom or not self.custom.band_dict:
            raise RuntimeError("No band dict registered for this device")

        bands = [BAND_OPTS_REV[key] for key in self.custom.band_dict.keys()
                    if key in BAND_OPTS_REV]
        return defer.succeed(sorted(bands))

    def configure_connection(self, options):
        """
        Configures the connection with the given parameters

        @type options: C{dict}
        """
        if not self.custom:
            raise RuntimeError("This device has no self.custom")

        if 'conn' not in options and 'band' not in options:
            raise RuntimeError("Unknown options %s" % options)

        try:
            d = self._configure_conn_mode(options)
        except RuntimeError:
            return self._configure_band_mode(options)
        else:
            if 'band' in options:
                d.addCallback(lambda _: self._configure_band_mode(options))
            return d

    def _configure_conn_mode(self, options):
        if not self.custom.conn_dict:
            raise RuntimeError("This device has no self.custom.conn_dict")

        if 'conn' in options:
            try:
                conn_mode = CONN_OPTS[options['conn']]
            except KeyError:
                msg = "Connection mode %s not found in CONN_OPTS"
                raise RuntimeError(msg % options['conn'])

            if conn_mode in self.custom.conn_dict:
                conn_str = self.custom.conn_dict[conn_mode]
                if conn_str:
                    return self.send_at(conn_str)
                else:
                    msg = "No AT string for connection mode %s"
                    raise RuntimeError(msg % conn_mode)
            else:
                msg = "Device %s doesn't have key %s in its conn_dict"
                raise RuntimeError(msg % (self, conn_mode))

        raise RuntimeError("conn not defined in %s" % options)

    def _configure_band_mode(self, options):
        if not self.custom.band_dict:
            raise RuntimeError("This device has no self.custom.band_dict")

        if 'band' in options:
            try:
                band = BAND_OPTS[options['band']]
            except KeyError:
                msg = "Band mode %s not found in BAND_OPTS"
                raise RuntimeError(msg % options['band'])

            if band in self.custom.band_dict:
                band_str = self.custom.band_dict[band]
                if band_str:
                    return self.send_at(band_str)
                else:
                    msg = "No AT string for band mode %s"
                    raise RuntimeError(msg % band)
            else:
                msg = "Device %s doesn't have key %s in its band_dict"
                raise RuntimeError(msg % (self, band))

        raise RuntimeError("band not defined in %s" % options)


class CDMAWrapper(BaseWrapper, CDMADeviceConnection):
    """
    Wrapper around L{CDMADeviceConnection}
    """

    def _enable_device(self):
        """
        Enables the device
        """
        d = self.reset_settings()
        d.addCallback(lambda ign: self.disable_echo())
        d.addCakkback(lambda ign: self.DeviceEnabled(self.device.udi))
        return d

    def get_esn(self):
        """
        Returns the device ESN
        """
        d = super(CDMAWrapper, self).get_esn()
        d.addCallback(lambda response: response[0].group('esn'))
        return d


class WCDMAWrapper(BaseWrapper, WCDMADeviceConnection):
    """
    Wrapper around L{WCDMADeviceConnection}

    Its main objective is to provide some error control on some operations
    and a cleaner API to deal with
    L{wader.common.protocol.WCDMADeviceConnection} results
    """

    def add_contact(self, contact):
        """
        Adds C{contact} to the SIM and returns the index where was stored

        @rtype: C{defer.Deferred}
        """
        name = from_u(contact.get_name())

        if 'UCS2' in self.device.sim.charset:
            name = pack_ucs2_bytes(name)

        # common arguments for both operations (name and number)
        args = [name, from_u(contact.get_number())]

        if contact.index:
            # contact.index is set, user probably wants to overwrite an
            # existing contact
            args.append(contact.index)
            d = super(WCDMAWrapper, self).add_contact(*args)
            d.addCallback(lambda _: contact.index)
            return d

        # contact.index is not set, this means that we need to obtain the
        # first slot free on the phonebook and then add the contact
        def get_next_id_cb(index):
            args.append(index)
            d2 = super(WCDMAWrapper, self).add_contact(*args)
            # now we just fake add_contact's response and we return the index
            d2.addCallback(lambda _: index)
            return d2

        d = super(WCDMAWrapper, self).get_next_contact_id()
        d.addCallback(get_next_id_cb)
        return d

    def add_sms(self, sms):
        """
        Adds C{sms} to the SIM archive
        """
        pdu_len = sms.get_pdu_len()
        d = super(WCDMAWrapper, self).add_sms(pdu_len, sms.get_pdu())
        d.addCallback(lambda resp: int(resp[0].group('id')))
        return d

    def change_pin(self, oldpin, newpin):
        """
        Changes PIN from C{oldpin} to C{newpin}
        """
        d = super(WCDMAWrapper, self).change_pin(oldpin, newpin)
        d.addCallback(lambda result: result[0].group('resp'))
        return d

    def check_pin(self):
        """
        Returns the SIM's auth state

        @raise ex.CMEErrorSIMPINRequired: Raised if SIM PIN is required
        @raise ex.CMEErrorSIMPUKRequired: Raised if SIM PUK is required
        @raise ex.CMEErrorSIMPUK2Required: Raised if SIM PUK2 is required
        """
        d = super(WCDMAWrapper, self).check_pin()
        d.addCallback(lambda result: result[0].group('resp'))
        def process_result(result):
            if result == 'READY':
                return result
            elif result == 'SIM PIN':
                raise ex.CMEErrorSIMPINRequired()
            elif result == 'SIM PUK':
                raise ex.CMEErrorSIMPUKRequired()
            elif result == 'SIM PUK2':
                raise ex.CMEErrorSIMPUK2Required()
            else:
                log.err("Unknown AUTH SIM STATE %s" % result)

        d.addCallback(process_result)
        d2 = defer.Deferred()
        d.chainDeferred(d2)
        return d2

    def delete_contact(self, index):
        d = super(WCDMAWrapper, self).delete_contact(index)
        d.addCallback(lambda result: result[0].group('resp'))
        return d

    def delete_sms(self, index):
        d = super(WCDMAWrapper, self).delete_sms(index)
        d.addCallback(lambda result: result[0].group('resp'))
        return d

    def disable_pin(self, pin):
        """
        Disables the PIN auth with C{pin}

        @rtype: C{defer.Deferred}
        """
        d = super(WCDMAWrapper, self).disable_pin(pin)
        d.addCallback(lambda result: result[0].group('resp'))
        return d

    def enable_pin(self, pin):
        """
        Enables the PIN auth with C{pin}

        @rtype: C{defer.Deferred}
        """
        d = super(WCDMAWrapper, self).enable_pin(pin)
        d.addCallback(lambda result: result[0].group('resp'))
        return d

    def find_contacts(self, pattern):
        """Returns a list of C{Contact} whose name matches pattern"""
        if 'UCS2' in self.device.sim.charset:
            pattern = pack_ucs2_bytes(pattern)
        d = super(WCDMAWrapper, self).find_contacts(pattern)
        d.addCallback(lambda matches: [process_contact_match(m)
                                        for m in matches])
        return d

    def get_available_charset(self):
        """
        Returns a list with the available character sets

        @rtype: C{defer.Deferred}
        """
        d = super(WCDMAWrapper, self).get_available_charset()
        d.addCallback(lambda resp: [match.group('lang') for match in resp])
        return d

    def get_charset(self):
        """
        Returns the current charset
        """
        d = super(WCDMAWrapper, self).get_charset()
        d.addCallback(lambda response: response[0].group('lang'))
        return d

    def get_contact_by_index(self, index):
        d = super(WCDMAWrapper, self).get_contact_by_index(index)
        d.addCallback(lambda match: process_contact_match(match[0]))
        return d

    def get_contacts(self):
        """Returns a list with all the contacts in the SIM"""
        d = super(WCDMAWrapper, self).get_contacts()
        def not_found_eb(failure):
            failure.trap(ex.CMEErrorNotFound, ex.ATError)
            return []

        d.addCallback(lambda matches:
                      [process_contact_match(m) for m in matches])
        d.addErrback(not_found_eb)
        return d

    def get_free_contact_ids(self):
        """Returns a deque with the contact ids not used"""
        def get_contacts_cb(used_ids):
            if not used_ids:
                # if there are no ids used, just return a list 1-sim size
                return deque(range(1, self.device.sim.size))

            # to compute the free ids, we just get XOR the set of 1 - sim size
            # with the set of busy_ids, that will leave a list with only the
            # ids not being used
            busy_ids = [contact.get_index() for contact in used_ids]
            free = set(range(1, self.device.sim.size)) ^ set(busy_ids)
            return deque(list(free))

        def get_contacts_eb(failure):
            failure.trap(ex.CMEErrorNotFound, ex.ATError)
            # some devices like to raise an exception when there are no
            # contacts
            return deque(range(1, self.device.sim.size))

        d = self.get_contacts()
        d.addCallbacks(get_contacts_cb, get_contacts_eb)
        return d

    def get_imei(self):
        """
        Returns the card's IMEI number

        @rtype: C{Deferred}
        """
        d = super(WCDMAWrapper, self).get_imei()
        d.addCallback(lambda response: response[0].group('imei'))
        return d

    def get_imsi(self):
        """
        Returns the SIM's IMSI number

        @rtype: C{Deferred}
        """
        d = super(WCDMAWrapper, self).get_imsi()
        d.addCallback(lambda response: response[0].group('imsi'))
        return d

    def get_netreg_status(self):
        """
        Returns a tuple with the network registration status

        +CREG: 0,0 - Not registered and not scanning for a GSM network
        +CREG: 0,1 - Registered on the "HOME" network of the SIM
        +CREG: 0,2 - Not registered but is scanning for a GSM network
        +CREG: 0,3 - Registration is denied (Manual attempt failed)
        +CREG: 0,5 - Registered on to another network (roaming).

        @rtype: C{Deferred}
        """
        d = super(WCDMAWrapper, self).get_netreg_status()
        def get_netreg_status(resp):
            # convert them to int
            return int(resp[0].group('mode')), int(resp[0].group('status'))

        d.addCallback(get_netreg_status)
        return d

    def get_network_info(self, process=True):
        """
        Returns the network info  (a.k.a AT+COPS?)

        The response will be a tuple as (OperatorName, ConnectionType) if
        it returns a (None, None) that means that some error occurred while
        obtaining the info. The class that requested the info should take
        care of insisting before this problem. This method will convert
        numeric network IDs to alphanumeric.

        @rtype: C{Deferred}
        """
        d = super(WCDMAWrapper, self).get_network_info()
        def get_net_info_cb(netinfo):
            """
            Returns a (Networname, ConnType) tuple

            It returns None if there's no info
            """
            if not netinfo:
                return None

            netinfo = netinfo[0]

            if netinfo.group('error'):
                # this means that we've received a response like
                # +COPS: 0 which means that we don't have network temporaly
                # we should raise an exception here
                raise ex.NetworkTemporalyUnavailableError

            status = int(netinfo.group('status'))
            conn_type = (status == 0) and 'GPRS' or '3G'
            netname = netinfo.group('netname')

            if netname in ['Limited Service',
                           pack_ucs2_bytes('Limited Service')]:
                raise ex.LimitedServiceNetworkError

            # netname can be in UCS2, as a string, or as a network id (int)
            if 'UCS2' in self.device.sim.charset and check_if_ucs2(netname):
                return unpack_ucs2_bytes(netname), conn_type
            else:
                # now can be either a string or a network id (int)
                try:
                    netname = int(netname)
                except ValueError:
                    # we got a string ID
                    return netname, conn_type

                # if we have arrived here, that means that the network id
                # is a five digit integer
                if not process:
                    return netname, conn_type

                # we got a numeric id, lets convert it
                from wader.common.persistent import net_manager
                network = net_manager.get_network_by_id(netname)
                if network:
                    return network.get_full_name(), conn_type

                return _('Unknown Network'), conn_type

        d.addCallback(get_net_info_cb)
        return d

    def get_network_names(self):
        """
        Returns a list of C{NetworkObject}s  (a.k.a AT+COPS=?)

        @rtype: C{Deferred}
        """
        d = self.set_charset("IRA")
        d.addCallback(lambda ign: super(WCDMAWrapper, self).get_network_names())
        def get_names_cb(resp):
            ret = [NetworkOperator(*match.groups()) for match in resp]
            d2 = self.set_charset("UCS2")
            d2.addCallback(lambda ign: ret)
            return d2

        d.addCallback(get_names_cb)
        return d

    def get_phonebook_size(self):
        """
        Returns the phonebook size of the SIM card

        @rtype: C{Deferred}
        """
        d = super(WCDMAWrapper, self).get_phonebook_size()
        d.addCallback(lambda resp: int(resp[0].group('size')))
        return d

    def get_pin_status(self):
        """
        Returns 1 if PIN auth is active and 0 if its not

        @rtype: C{Deferred}
        """
        def pinreq_errback(failure):
            failure.trap(ex.CMEErrorSIMPINRequired)
            return 1

        def aterror_eb(failure):
            failure.trap(ex.ATError)
            # return the failure or wont work
            return failure

        d = super(WCDMAWrapper, self).get_pin_status()
        d.addCallback(lambda response: int(response[0].group('status')))
        d.addErrback(pinreq_errback)
        d.addErrback(aterror_eb)

        return d

    def get_roaming_ids(self):
        """
        Returns the network ids stored in the SIM to roam (a.k.a AT+CPOL?)

        @rtype: C{Deferred}
        """
        d = super(WCDMAWrapper, self).get_roaming_ids()
        def get_roaming_ids_cb(raw):
            return [BasicNetworkOperator(obj.group('netid')) for obj in raw]

        d.addCallback(get_roaming_ids_cb)
        return d

    def get_sms(self):
        """
        Returns a list of Message objects with all the SMS in the SIM card
        """
        d = super(WCDMAWrapper, self).get_sms()
        def get_all_sms_cb(messages):
            sms_list = []
            for rawsms in messages:
                # Message obj
                try:
                    sms = pdu_to_message(rawsms.group('pdu'))
                    sms.index = int(rawsms.group('id'))
                    sms.where = int(rawsms.group('storedat'))
                    sms_list.append(sms)
                except ValueError:
                    log.err(ex.MalformedSMSError,
                            "Malformed PDU: %s" % rawsms.group('pdu'))
            return sms_list

        d.addCallback(get_all_sms_cb)
        return d

    def get_sms_by_index(self, index):
        """
        Returns a C{Message} object representing the SMS at C{index}

        @rtype: C{Deferred}
        """
        d = super(WCDMAWrapper, self).get_sms_by_index(index)
        def get_sms_cb(rawsms):
            try:
                sms = pdu_to_message(rawsms[0].group('pdu'))
                sms.where = int(rawsms[0].group('storedat'))
                sms.index = index
            except IndexError:
                # handle bogus CMTI notifications, see #180
                return None

            return sms

        d.addCallback(get_sms_cb)
        return d

    def get_sms_format(self):
        """
        Returns 1 if SMS format is text and 0 if SMS format is PDU
        """
        d = super(WCDMAWrapper, self).get_sms_format()
        d.addCallback(lambda response: int(response[0].group('format')))
        return d

    def get_smsc(self):
        """
        Returns the SMSC number stored in the SIM

        @rtype: C{Deferred}
        """
        d = super(WCDMAWrapper, self).get_smsc()
        def get_smsc_cb(response):
            try:
                smsc = response[0].group('smsc')
                if not smsc.startswith('+'):
                    if ('UCS2' in self.device.sim.charset
                            and check_if_ucs2(smsc)):
                        # the smsc is in UCS2 format
                        smsc = from_u(unpack_ucs2_bytes(smsc))

                return smsc
            except KeyError, e:
                raise ex.CMEErrorNotFound()

        d.addCallback(get_smsc_cb)
        return d

    def get_used_contact_ids(self):
        """Returns a list with the used contact ids"""
        def callback(contacts):
            if not contacts:
                return []

            return [contact.get_index() for contact in contacts]

        def errback(failure):
            failure.trap(ex.CMEErrorNotFound, ex.ATError)
            return []

        d = self.get_contacts()
        d.addCallback(callback)
        d.addErrback(errback)
        return d

    def get_used_sms_ids(self):
        """Returns a list with used SMS ids in the SIM card"""
        d = self.get_sms()
        def process_all_sms(smslist):
            if not smslist:
                return []

            return [sms.get_index() for sms in smslist]

        def errback(failure):
            failure.trap(ex.CMEErrorNotFound, ex.ATError)
            return []

        d.addCallback(process_all_sms)
        d.addErrback(errback)
        return d

    def save_sms(self, sms):
        """
        Returns the index where C{sms} was stored
        """
        pdu_len, pdu = message_to_pdu(sms)
        d = super(WCDMAWrapper, self).save_sms(pdu, pdu_len)
        d.addCallback(lambda response: response[0].group('index'))
        return d

    def send_pin(self, pin):
        """
        Sends C{pin} and returns the auth result

        Most devices need some time to settle after a successful auth
        it is the caller's responsability to give at least 15 seconds
        to the device to settle, this time varies from device to device
        """
        d = super(WCDMAWrapper, self).send_pin(pin)
        d.addCallback(lambda response: response[0].group('resp'))
        return d

    def send_sms(self, sms):
        """Sends C{sms} and returns the index"""
        pdu_len, pdu = message_to_pdu(sms)
        d = super(WCDMAWrapper, self).send_sms(pdu, pdu_len)
        d.addCallback(lambda response: int(response[0].group('index')))
        return d

    def send_sms_from_storage(self, index):
        """Sends the SMS stored at C{index} and returns the new index"""
        d = super(WCDMAWrapper, self).send_sms_from_storage(index)
        d.addCallback(lambda response: int(response[0].group('index')))
        return d

    def set_charset(self, charset):
        """Sets the SIMs charset to C{charset}"""
        d = super(WCDMAWrapper, self).set_charset(charset)
        d.addCallback(lambda ignored: self.device.sim.set_charset(charset))
        return d

    def set_sms_format(self, format=0):
        """Sets PDU mode or text mode in the SIM"""
        d = super(WCDMAWrapper, self).set_sms_format(format)
        d.addCallback(lambda response: response[0].group('resp'))
        return d

    def set_smsc(self, smsc):
        """Sets the SIMS's SMSC number to C{smsc}"""
        if 'UCS2' in self.device.sim.charset:
            smsc = pack_ucs2_bytes(smsc)
        d = super(WCDMAWrapper, self).set_smsc(smsc)
        d.addCallback(lambda response: response[0].group('resp'))
        return d

    # some high-level methods exported over DBus

    def _enable_device(self):
        """
        I check the auth state of a device and will try to initialize it
        """
        from wader.common.statem.auth import DBusCollabMixin
        from wader.utils.utilities import MixIn
        authklass = self.device.custom.authklass
        MixIn(authklass, DBusCollabMixin)
        authsm = authklass(self.device, DBusCollaboratorFactory)
        d = authsm.start_auth()
        # if auth is ready, the device will initialize straight away
        # if auth aint ready the callback chain wont be executed and
        # will just return the given exception
        d.addCallback(self.device.initialize)
        d.addCallback(self.device.wrapper.daemons.start_daemons)
        d.addCallback(lambda ign: self.DeviceEnabled(self.device.udi))
        return d

    def _check_initted_device(self, result):
        """
        Upon successful auth over DBus I'll check if the device was initted
        """
        if self.device.sim and self.device.sim.initted:
            log.msg("Device was already initted, just returning orig result")
            return result

        DELAY = self.device.custom.authklass.DELAY
        log.msg("Giving %d seconds to settle the device, waiting..." % DELAY)

        d = defer.Deferred()
        reactor.callLater(DELAY, self._init_device, d, result)
        return d

    def _init_device(self, d2, result):
        d = self.device.initialize()
        d.addCallback(self.device.wrapper.daemons.start_daemons)
        d.addCallback(lambda _: result)
        d.chainDeferred(d2)


class BasicNetworkOperator(object):
    """
    A Network operator with a netid
    """
    def __init__(self, netid=None):
        super(BasicNetworkOperator, self).__init__()
        self.netid = netid

    def __repr__(self):
        args = (id(self), str(self.netid))
        return '<BasicNetworkOperator at 0x%x %s>' % args

    def __eq__(self, o):
        if isinstance(o, BasicNetworkOperator):
            return self.netid == o.netid

        raise ValueError("You cannot reliably compare %s with %s" % (self, o))

    def __ne__(self, o):
        return not self.__eq__(o)


class NetworkOperator(BasicNetworkOperator):
    """
    I represent a network operator on a mobile network
    """
    def __init__(self, stat, long_name, short_name, netid, rat):
        super(NetworkOperator, self).__init__(int(netid))
        self.stat = int(stat)
        self.long_name = from_ucs2(long_name)
        self.short_name = from_ucs2(short_name)
        self.rat = int(rat)

    def __repr__(self):
        args = (self.long_name, self.netid)
        return '<NetworkOperator "%s" netid: %d>' % args

