# -*- coding: utf-8 -*-
# Copyright (C) 2006-2008  Vodafone España, S.A.
# Copyright (C) 2008  Warp Networks, S.L.
# Author:  Pablo Martí
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""Plugin system for WaderCdfL2"""

import os
from pytz import timezone

from zope.interface import implements

from twisted.python import log
from twisted.plugin import IPlugin, getPlugins

import wader.common.exceptions as ex
import wader.common.interfaces as interfaces
from wader.common.sim import SIMBaseClass


def flatten_list(x):
    """
    Flattens C{x} into a single list
    """
    result = []
    for el in x:
        if hasattr(el, "__iter__") and not isinstance(el, basestring):
            result.extend(flatten_list(el))
        else:
            result.append(el)
    return result


class DevicePlugin(object):
    """Base class of all DevicePlugins"""

    implements(IPlugin, interfaces.IDevicePlugin)
    # at what speed should we talk with this device?
    baudrate = 115200
    # Class that will initialize the SIM, by default SIMBaseClass
    simklass = SIMBaseClass
    # Response of AT+CGMM
    __remote_name__ = ""
    # instance of a custom adapter class if device needs customization
    custom = None
    # attributes to track some nasty cards
    dodgy = False
    dodgy_ports = ()
    dialer = 'default'

    def __init__(self):
        super(DevicePlugin, self).__init__()
        # control port
        self.cport = None
        # data port
        self.dport = None
        # sim instance
        self.sim = None
        # SerialPort reference
        self.sport = None
        # serial connection reference
        self.sconn = None

    def __repr__(self):
        args = [self.__class__.__name__, self.dport]
        if self.cport:
            args.append(self.cport)
            return "<%s  dport: %s  cport: %s>" % tuple(args)

        return "<%s  dport: %s>" % tuple(args)


    def close(self):
        """
        Closes the plugin and frees all the associated resources
        """
        try:
            if self.sconn:
                self.sconn.remove_from_connection()
                self.sconn.transport.unregisterProducer()
        except LookupError:
            # the object wasn't exported at all
            pass

        except:
            log.err()
        try:
            if self.sport:
                self.sport.connectionLost("Closing connection")
                self.sport.loseConnection("Bye!")
        except:
            log.err()

    def initialize(self, arg=None):
        """
        Initializes the SIM

        @rtype: L{twisted.internet.defer.Deferred}
        """
        self.sim = self.simklass(self.sconn)
        return self.sim.initialize()

    def has_two_ports(self):
        """
        Returns C{True} if the device sports two ports
        """
        try:
            return self.cport != None
        except AttributeError:
            return False

    def patch(self, other):
        """
        Patch myself in-place with the settings of another plugin
        """
        if not isinstance(other, DevicePlugin):
            raise ValueError("Cannot patch myself with a %s" % type(other))

        self.cport = other.cport
        self.dport = other.dport
        self.baudrate = other.baudrate


class DBusDevicePlugin(DevicePlugin):
    """
    Base class from which all the DBusDevicePlugins should inherit from
    """
    implements(IPlugin, interfaces.IDBusDevicePlugin)

    __properties__ = {}

    def __init__(self):
        super(DBusDevicePlugin, self).__init__()
        self.udi = None
        # onyl used in devices that like to share ids, like
        # huawei's exxx family. It should have at least a
        # 'default' key mapping to a safe device that can be
        # used to identify the rest of the family
        self.mapping = {}

    def patch(self, other):
        super(DBusDevicePlugin, self).patch(other)
        self.udi = other.udi


class RemoteDevicePlugin(DevicePlugin):
    """
    Base class from which all the RemoteDevicePlugins should inherit from
    """
    implements(IPlugin, interfaces.IRemoteDevicePlugin)


class UnknownDevicePlugin(DevicePlugin):
    """
    UnknownDevicePlugin
    """
    implements(IPlugin, interfaces.IRemoteDevicePlugin)


BASE_PATH_DICT = {}

class OSPlugin(object):
    """Base class from which all the OSPlugins should inherit from"""
    implements(IPlugin, interfaces.IOSPlugin)
    abstraction = BASE_PATH_DICT
    customization = None
    dialer = None
    os_name = None
    os_version = None
    manage_secrets = True
    privileges_needed = None

    def __init__(self):
        super(OSPlugin, self).__init__()

    def are_privileges_needed(self):
        """
        Sets C{self.privileges_needed} to True if pppd is suid

        This should take into account SELinux stuff
        """
        if not os.geteuid():
            # if we're running the bundled version or we are root
            return False

        return False

    def get_timezone(self):
        raise NotImplementedError()

    def get_tzinfo(self):
        zone = self.get_timezone()
        try:
            return timezone(zone)
        except:
            # we're not catching this exception because some dated pytz
            # do not include UnknownTimeZoneError, if get_tzinfo doesn't works
            # we just return None as its a valid tzinfo and we can't do more
            return None

    def get_iface_stats(self, iface):
        raise NotImplementedError()

    def is_valid(self, os_info=None):
        raise NotImplementedError()

    def initialize(self):
        self.privileges_needed = self.are_privileges_needed()
        if self.customization:
            self.abstraction.update(self.customization)


class NotificationPlugin(object):
    """Base class from which all NotificationPlugins should inherit from"""
    implements(IPlugin, interfaces.INotificationPlugin)
    klass = None # reference to the class that we are interested in

    def __init__(self):
        super(NotificationPlugin, self).__init__()

    def on_notification_received(self, wrapper, notification):
        raise NotImplementedError()


import wader.common.plugins
class PluginManager(object):
    """I manage WaderCdfL's plugins"""

    @classmethod
    def get_plugins(cls, interface=IPlugin, package=wader.common.plugins):
        return getPlugins(interface, package)

    @classmethod
    def regenerate_cache(cls):
        log.msg("PluginManager: Regenerating plugin cache...")
        list(getPlugins(IPlugin, package=wader.common.plugins))

    @classmethod
    def get_plugin_by_remote_name(cls, name,
                                  interface=interfaces.IDevicePlugin):
        """
        Get a plugin by its remote name

        @raise ex.UnknownPluginNameError: When we don't know about the plugin
        """
        for plugin in cls.get_plugins(interface, wader.common.plugins):
            try:
                if plugin.__remote_name__ == name:
                    return plugin
            except AttributeError:
                # if Plugin.__class__ != DevicePlugin, __remote_name__ not def
                continue

        raise ex.UnknownPluginNameError(name)

    @classmethod
    def get_plugin_by_vendor_product_id(cls, product_id, vendor_id):
        log.msg("get_plugin_by_id called with 0x%X and 0x%X" % (product_id,
                                                            vendor_id))
        for plugin in cls.get_plugins(interfaces.IDBusDevicePlugin):
            props = flatten_list(plugin.__properties__.values())
            if int(product_id) in props and int(vendor_id) in props:
                if not plugin.mapping:
                    # regular plugin
                    return plugin

                # device has multiple personalities...
                # this will just return the default plugin for
                # the mapping, we keep a reference to the mapping
                # once the device is properly identified by
                # wader.common.hardware.base::identify_device
                _plugin = plugin.mapping['default']()
                _plugin.mapping = plugin.mapping
                return _plugin

        return None

