# -*- coding: utf-8 -*-
# Copyright (C) 2006-2008  Vodafone España, S.A.
# Author:  Pablo Martí
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Profile manager

I manage mobile profiles in the system
"""

import os

import dbus
from dbus.service import Object, BusName, method

from wader.common.consts import (MOBILE_PROFILES, WADER_PROFILES_SERVICE,
                                 WADER_PROFILES_OBJECT, WADER_PROFILES_OPATH,
                                 PRF_INTFACE)
from wader.common.configbase import MobileProfile
from wader.common.persistent import net_manager
from wader.utils.utilities import natsort
import wader.common.exceptions as ex


class ProfileManager(Object):
    """
    I manage mobile profiles
    """

    def __init__(self, path=MOBILE_PROFILES):
        name = BusName(WADER_PROFILES_SERVICE, bus=dbus.SessionBus())
        super(ProfileManager, self).__init__(bus_name=name,
                                            object_path=WADER_PROFILES_OBJECT)
        self.path = path

    def close(self):
        self.remove_from_connection()

    def create_profile(self, opts):
        """
        Creates a new mobile profile from C{opts}

        @type opts: dict
        """
        ids = self.get_profile_ids()
        if not ids:
            last = str(0)
        else:
            natsort(ids)
            last = str(int(ids[-1]) + 1)

        opath = os.path.join(WADER_PROFILES_OPATH, last)
        profpath = os.path.join(MOBILE_PROFILES, last)
        profile = MobileProfile.from_dict(opts, opath, profpath)
        profile.write()

        return profile

    def delete_profile(self, profile):
        """
        Deletes C{profile}

        @raise wader.common.exceptions.ProfileInUseError: If the profile to
        delete is the active profile
        @raise OSError: Raised if we don't have permissions to delete profile
        @raise IOError: Raised if profile doesn't exists or any other IO error
        """
        profile.close()

        try:
            os.unlink(profile.path)
        except (IOError, OSError):
            raise

    def edit_profile(self, profile, opts):
        """
        Edits C{profile} from the C{opts} dict
        """
        self._configure_profile(profile.name, opts)
        # XXX
        #if profile == self.config.current_profile:
            #    self._configure_connection(profile)

    def get_profiles(self):
        ids = self.get_profile_ids()
        profiles = []
        for _id in ids:
            opath = os.path.join(WADER_PROFILES_OPATH, _id)
            profpath = os.path.join(self.path, _id)
            profiles.append(MobileProfile(opath, profpath))

        return profiles

    def get_profile_opaths(self):
        """
        Returns all the stored profiles as a list of C{MobileProfile} objects
        """
        return [os.path.join(WADER_PROFILES_OPATH, name) for name in
                os.listdir(self.path)]

    def get_profile_ids(self):
        return [os.path.basename(name) for name in os.listdir(self.path)]


class DBusProfileManager(ProfileManager):
    """
    DBus layer over C{ProfileManager}
    """
    def __init__(self, path=MOBILE_PROFILES):
        super(DBusProfileManager, self).__init__(path)
        self.profiles = {}
        self.cache_profiles()

    def cache_profiles(self):
        for profile in self.get_profiles():
            self.profiles[profile.opath] = profile

    @method(PRF_INTFACE, in_signature='a{ss}', out_signature='o')
    def CreateProfile(self, opts):
        profile = self.create_profile(opts)
        self.profiles[profile.opath] = profile
        return profile.opath

    @method(PRF_INTFACE, in_signature='s', out_signature='o')
    def CreateProfileFromIMSI(self, imsi):
        network = net_manager.get_network_by_id(imsi)
        if not network:
            raise ex.ProfileNotFoundError()

        opts = {}
        opts['name'] = network.name
        opts['username'] = network.username
        opts['password'] = network.password
        opts['apn'] = network.apn
        opts['connection'] = '3GPREF'
        opts['dns1'] = network.dns1
        opts['dns2'] = network.dns2

        profile = self.create_profile(opts)
        self.profiles[profile.opath] = profile
        return profile.opath

    @method(PRF_INTFACE, in_signature='oa{ss}', out_signature='')
    def ModifyProfile(self, opath, opts):
        if not opath in self.profiles:
            raise RuntimeError("Profile %s does not exists" % opath)

        profile = self.profiles[opath]
        profile.patch(opts)


    @method(PRF_INTFACE, in_signature='o', out_signature='')
    def DeleteProfile(self, opath):
        if not opath in self.profiles:
            raise RuntimeError("Profile %s does not exists" % opath)

        profile = self.profiles[opath]
        del self.profiles[opath]
        self.delete_profile(profile)

    @method(PRF_INTFACE, in_signature='', out_signature='ao')
    def GetProfiles(self):
        return self.profiles.keys()




