# -*- coding: utf-8 -*-
# Copyright (C) 2004 Paul Hardwick paul@peck.org.uk
# Copyright (C) 2008 Warp Networks S.L.
# Imported for the wader project on 5 June 2008 by Pablo Martí
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# $Id: pdu.py 72 2004-09-24 15:31:51Z paul $
#

from wader.contrib.pdu_sms import mapgsm

SEVENBIT = 0x00
EIGHTBIT = 0x04
UNICODE =  0x08

# extracted from the Phebe project
GSM_DEFAULT_TABLE = u"\
@£$¥èéùìòÇ\nØø\rÅå\
Δ_ΦΓΛΩΠΨΣΘΞ ÆæßÉ\
 !\"#¤%&'()*+,-./\
0123456789:;<=>?\
¡ABCDEFGHIJKLMNO\
PQRSTUVWXYZÄÖÑÜ§\
¿abcdefghijklmno\
pqrstuvwxyzäöñüà\
\f^{}\\[~]|€"

def is_7bit_text(text):
    for c in text:
        if c not in GSM_DEFAULT_TABLE:
            return False
    return True

class PDU(object):

    def __init__(self, pduformat=None):
        """
        Create a C{PDU} with encoding specified by C{pduformat}

        If C{pduformat} is None, it will try to infer the encoding itself

        pdu.SEVENBIT = 7-bit GSM text  (160 char limit)
        pdu.EIGHTBIT = 8-bit code - no manipulation (140 char limit)
        pdu.UNICODE = each char is two byte unicode  (70 char limit)
        """
        super(PDU, self).__init__()
        self.pduformat = pduformat
        self.pdusms = ""
        self.ignoreUDHI = True

    def _find_out_pduformat(self, text):
        if is_7bit_text(text):
            self.pduformat = SEVENBIT
        else:
            self.pduformat = UNICODE

    def encode_pdu(self, number, text, csca='',
                    request_status=False, msgref=0, msgvp=0xaa):
        """
        Returns a tuple of message length and PDU

        csca = SMSC number
        request_status = True, False
        msgref = currently ignored integer 0-255
        msgvp = relative validity period as per ETSI (0xff == no validity)
        Message Encoding format defined by pduformat attribute
        SMSC can be ommited if phone can use internal defaults
        """
        if not self.pduformat:
            self._find_out_pduformat(text)

        if self.pduformat == SEVENBIT:
            messagetext = mapgsm.u2g(text)

        if not request_status:
            smssubmit = 0x01
        else:
            smssubmit = 0x21

        if msgvp != 0xFF:
            smssubmit = smssubmit | 0x10

        tpmessref = msgref & 0xFF
        tppid = 0x00  #Standard SMS protocol
        # Data coding scheme 00 = standard class 0 (7-bit) msg.
        tpdcs = 0x00 | self.pduformat
        # Unlimited validity period
        tpvalidity = msgvp & 0xFF
        # convert csca to correct format
        smscpdu = self._get_smsc_pdu(csca)
        # make a note of smscpdu length
        smcslength = len(smscpdu)
        # convert to pdu format
        recippdu = self._get_phone_pdu(number)
        # Convert message based on pduformat setting
        messagepdu = self._convert_message(text)

        # Create final PDU self.message array
        self.pdusms = smscpdu
        self.pdusms = self.pdusms + chr(smssubmit) + chr(tpmessref)
        self.pdusms = self.pdusms + recippdu
        self.pdusms = self.pdusms + chr(tppid) + chr(tpdcs)
        #if Validity flag set then include validity field
        if smssubmit & 0x18:
            self.pdusms = self.pdusms + chr(tpvalidity)

        self.pdusms = self.pdusms + messagepdu
        pdu = ''
        for n in self.pdusms:
            pdu = pdu + "%02x" % ord(n)

        return (len(pdu)/2 - smcslength), pdu.upper()

    def _unpack_msg(self, pdu, limit=160):
        """
        Unpacks C{pdu} into 7-bit characters and returns the decoded string
        """
        # Taken/modified from Dave Berkeley's pysms package
        count = 0
        last = 0
        result = []
        for i in range(0, len(pdu), 2):
            byte = int(pdu[i:i+2], 16)
            mask = 0x7F >> count
            out = ((byte & mask) << count) + last
            last = byte >> (7 - count)
            result.append(chr(out))
            if limit and (len(result) >= limit):
                break
            if count == 6:
                result.append(chr(last))
                last = 0

            count = (count + 1) % 7
        return ''.join(result)

    def decode_pdu(self, pdu):
        """
        Decodes a complete SMS pdu and returns a tuple strings
        sender,  # Senders number
        datestr, # GSM format date string
        msg,     # The actual msg less any header in UCS2 format
        ref,     # Msg reference  (from SMSC)
        cnt,     # Number of msg in the sequence (concat msgs)
        seq      # Sequence number of the msg (concat msgs)
        fmt      # Format of recieved msg
        """
        pdu = pdu.upper()
        ptr = 0
        #Service centre address
        smscl = int(pdu[ptr:ptr+2], 16) * 2 # number of digits
        smscertype = pdu[ptr+2:ptr+4]
        smscer = pdu[ptr+4:ptr+smscl+2]
        smscer = smscer.replace('F', '')
        smscer = list(smscer)
        for n in range(1, len(smscer), 2):
            smscer[n-1], smscer[n] = smscer[n], smscer[n-1]

        smscer = ''.join(smscer)
        if smscertype == '91':
            smscer = '+' + smscer
        csca = smscer

        ptr = ptr + 2 + smscl
        #1 byte(octet) == 2 char
        #Message type TP-MTI bits 0,1
        #More messages to send/deliver bit 2
        #Status report request indicated bit 5
        #User Data Header Indicator bit 6
        #Reply path set bit 7
        #1st octet position == smscerlen+4

        FO = int(pdu[ptr:ptr+2], 16)
        #is this a concatenated msg?
        if (FO & 0x40) or self.ignoreUDHI:
            testheader = True
        else:
            testheader = False

        ptr = ptr + 2
        sndlen = int(pdu[ptr:ptr+2], 16)
        if sndlen % 2:
            sndlen += 1
        sndtype = pdu[ptr+2:ptr+4]
        # Extract Phone number of sender
        sender = pdu[ptr+4:ptr+4+sndlen]
        sender = sender.replace('F', '')
        sender = list(sender)
        for n in range(1, len(sender), 2):
            sender[n-1], sender[n] = sender[n], sender[n-1]
        sender = ''.join(sender)
        sender = sender.strip()
        if sndtype == '91':
            sender = '+' + sender

        ptr = ptr + 4 + sndlen
        # 1byte (octet) = 2 char
        # 1 byte TP-PID (Protocol Identtifier
        PID = int(pdu[ptr:ptr+2], 16)
        ptr = ptr + 2
        # 1 byte TP-DCS (Data Coding Scheme)
        DCS = int(pdu[ptr:ptr+2], 16)
        fmt = SEVENBIT
        if DCS & (EIGHTBIT|UNICODE) == 0:
            fmt = SEVENBIT
        elif DCS & EIGHTBIT:
            fmt = EIGHTBIT
        elif DCS & UNICODE:
            fmt = UNICODE

        # Get date stamp
        ptr = ptr + 2
        date = list(pdu[ptr:ptr+14])
        for n in range(1, len(date), 2):
            date[n-1], date[n] = date[n], date[n-1]
            datestr = "%s%s/%s%s/%s%s %s%s:%s%s:%s%s" % tuple(date[0:12])

        # Now get message body
        ptr = ptr + 14
        msgl = int(pdu[ptr:ptr+2], 16)
        msg = pdu[ptr+2:]
        # check for header
        cnt = 0
        seq = 0
        ref = 0
        headlen = 0
        if testheader:
            if msg[2:4] == "00": # found header for concat message
                headlen = (int(msg[0:2], 16) + 1) * 8
                subheadlen = int(msg[4:6], 16)
                ref = int(msg[6:8], 16)
                cnt = int(msg[8:10], 16)
                seq = int(msg[10:12], 16)
                if fmt == SEVENBIT:
                    while headlen % 7:
                        headlen += 1
                    headlen = headlen / 7
        if fmt == SEVENBIT:
            msg = self._unpack_msg(msg)[headlen:msgl]
            msg = mapgsm.g2u(msg)[:]
        elif fmt == EIGHTBIT:
            msg = ''.join([chr(int(msg[x:x+2], 16))
                                    for x in range(0, len(msg), 2)])
            msg = unicode(msg[headlen:], 'latin-1')
        elif fmt == UNICODE:
            msg = u''.join([unichr(int(msg[x:x+4], 16))
                            for x in range(0, len(msg), 4)])
        return sender, datestr, msg.strip(), csca, ref, cnt, seq, fmt

    def decode_cds_pdu(self, pdu):
        """
        Decodes a complete CDS pdu and returns   a tuple    strings
        sender,  # Senders number
        datestr, # GSM format date string
        msg,     # The actual msg less any header in utf-8 format
        ref,     # Msg reference  (from SMSC)
        cnt,     # Number of msg in the sequence (concat msgs)
        seq      # Sequence number of the msg (concat msgs)
        """

        # SMSC 6chars SENDER DATERECEIVED DATESTAMP
        # Status message PDU
        smscerlen = int(pdu[1], 16) * 2 # number of digits
        # print smscerlen
        smscertype = pdu[2:4]
        smscer = pdu[4:4 + smscerlen]
        smscer = smscer.replace('f',' ')
        smscer = smscer.replace('F',' ')
        smscer = list(smscer)
        for n in range(1, len(smscer), 2):
            smscer[n-1], smscer[n] = smscer[n], smscer[n-1]
        smscer = ''.join(smscer)
        if smscertype == '91':
            smscer = '+' + smscer
        msgref = int(pdu[smscerlen+4:smscerlen+6], 16)
        sndpos = 6 + smscerlen
        sndlen = int(pdu[sndpos+1], 16)
        # print sndlen
        sndtype = pdu[sndpos+2:sndpos+4]
        # print sndtype
        sender = pdu[sndpos+4:sndpos+4+sndlen]
        sender = sender.replace('f',' ')
        sender = sender.replace('F',' ')
        sender = list(sender)
        for n in range(1, len(sender), 2):
            sender[n-1], sender[n] = sender[n], sender[n-1]
        sender = ''.join(sender)
        if sndtype == '91':
            sender = '+' + sender
        # print sender
        mpos = -30
        # print pdu[mpos:-16]
        # msg = pdu[mpos:-16]
        # print msg
        date = list(pdu[mpos:-16])
        for n in range(1, len(date), 2):
            date[n-1], date[n] = date[n], date[n-1]
            msgstr = "%s%s/%s%s/%s%s %s%s:%s%s:%s%s" % (tuple(date[0:12]))

        date = list(pdu[-16:])
        for n in range(1, len(date), 2):
            date[n-1], date[n] = date[n], date[n-1]
            datestr = "%s%s/%s%s/%s%s %s%s:%s%s:%s%s" % (tuple(date[0:12]))

        status = int("0x%s" % pdu[-2:], 16)
        return sender, datestr, msgstr.strip(), status, msgref

    def _cleanNumber(self, number):
        """
        Returns clean number

        Makes sure numbers are stripped of unwanted characters
        """
        number = number.strip()
        number = number.replace(' ','')
        return number

    def _get_phone_pdu(self, number):
        """
        Converts C{number} to pdu format
        """
        number = self._cleanNumber(number)
        ptype = 129
        if number[0] == '+':
            number = number[1:]
            ptype = 145
        pl = len(number)
        if len(number) % 2:
            number = number + 'F'
        ps = chr(ptype)
        for n in range(0, len(number), 2):
            num = number[n+1] + number[n]
            ps = ps + chr(int(num, 16))
        ps = chr(pl) + ps
        return ps

    def _get_smsc_pdu(self, number):
        """
        Converts SMSC number string to expected pdu format for SMSC number
        """
        if not len(number.strip()):
            return chr(0)
        number = self._cleanNumber(number)
        ptype = 129
        if number[0] == '+':
            number = number[1:]
            ptype = 145
        if len(number) % 2:
            number = number + 'F'
        ps = chr(ptype)
        for n in range(0, len(number), 2):
            num = number[n+1] + number[n]
            ps = ps + chr(int(num, 16))
        pl = len(ps)
        ps = chr(pl) + ps
        return ps

    def _convert_message(self, message):
        """
        Converts 7-bit and Unicode message to correct format

        Does nothing with 8-bit  except calc length at front
        """
        pdu = ''
        if self.pduformat == SEVENBIT:
            txt = message
            tl = len(txt)
            txt = txt+'\x00'
            msgl = len(txt)*7/8
            op = [-1]*msgl
            c = 0
            shift = 0
            for n in range(msgl):
                if shift == 6:
                    c = c + 1
                shift = n % 7
                lb = ord(txt[c]) >> shift
                hb = (ord(txt[c+1]) << (7-shift) & 255)
                op[n] = lb + hb
                c += 1
            pdu = chr(tl) + ''.join([chr(x) for x in op])

        elif self.pduformat == EIGHTBIT:
            mlen = len(message)
            pdu = chr(mlen) + message

        elif self.pduformat == UNICODE:
            nmesg = ''
            for n in message:
                nmesg = nmesg + chr(ord(n) >> 8) + chr(ord(n) & 0xFF)
            mlen = len(message) * 2
            pdu = chr(mlen) + nmesg

        return pdu

