# -*- coding: utf-8 -*-
# Copyright (C) 2006-2008  Vodafone España, S.A.
# Copyright (C) 2008  Warp Networks, S.L.
# Author:  Pablo Martí
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""Unittests for DBus exported methods"""

import os

import dbus
import dbus.mainloop.glib
gloop = dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)

from twisted.internet import defer
from twisted.python import log
from twisted.trial import unittest

from wader.common.configbase import WaderConfigBase
from wader.common.consts import (WADER_SERVICE, PIN_INTFACE, CTS_INTFACE,
                               SMS_INTFACE, NET_INTFACE, CRD_INTFACE)
import wader.common.exceptions as ex
from wader.common.hardware import HardwareManager
from wader.common.startup import attach_serial_protocol
from wader.common.middleware import NetworkOperator
from wader.common.persistent import Contact

PATH = '/tmp/settings.conf'
temp_udi = "/wader/temp/udi"
conf = None

if os.path.exists(PATH):
    conf = WaderConfigBase(PATH)


class TestDBusExportedMethods(unittest.TestCase):
    """
    Tests for the exported methods over D-Bus
    """
    def setUpClass(self):
        self.sconn = None
        self.serial = None
        self.device = None
        hw = HardwareManager()
        try:
            d = hw.get_devices()
            def get_device_cb(devices):
                self.device = attach_serial_protocol(devices[0], temp_udi)
                self.sconn = self.device.sconn
                self.sport = self.device.sport
                d2 = self.device.initialize()
                d2.addCallback(self._get_remote_obj)
                d2.addCallback(lambda ign: self.sconn.delete_all_sms())
                d2.addCallback(lambda ign: self.sconn.delete_all_contacts())
                return d2

            d.addCallback(get_device_cb)
            return d
        except:
            log.err()

    def tearDownClass(self):
        return self.device.close()

    def _get_remote_obj(self, ignored=None):
        bus = dbus.SessionBus(mainloop=gloop)
        try:
            self.remote_obj = bus.get_object(WADER_SERVICE, temp_udi)
        except:
            log.err()
        return True

    def test_AddContact(self):
        """
        Test that es.warp.Wader.Contacts.Add works
        """
        c = Contact("Jhonny", "+435443434343")
        d = defer.Deferred()
        def remote_add_contact_cb(index):
            d2 = self.sconn.get_contact_by_index(index)
            d2.addCallback(lambda contact: self.assertEqual(contact, c))
            d2.addCallback(lambda _: self.sconn.delete_contact(index))
            d2.addCallback(lambda _: d.callback(True))

        def remote_add_contact_eb(failure):
            log.err(failure)

        self.remote_obj.Add(c.name, c.number,
                            dbus_interface=CTS_INTFACE,
                            reply_handler=remote_add_contact_cb,
                            error_handler=remote_add_contact_eb)
        return d

    def test_DeleteContact(self):
        """
        Test that es.warp.Wader.Contacts.Delete works
        """
        def remove_contact(index):
            d = defer.Deferred()
            def remote_del_contact_cb():
                d2 = self.sconn.get_contact_by_index(index)
                d3 = self.failUnlessFailure(d2, ex.ATError)
                d3.chainDeferred(d)

            def remote_del_contact_eb(failure):
                log.err(failure)

            self.remote_obj.Delete(index,
                                   dbus_interface=CTS_INTFACE,
                                   reply_handler=remote_del_contact_cb,
                                   error_handler=remote_del_contact_eb)

            return d

        c = Contact("OM", "+215443434343")
        d = self.sconn.add_contact(c)
        d.addCallback(remove_contact)
        return d

    def test_DisablePIN(self):
        """
        Test that es.warp.Wader.PIN.Disable works
        """
        pin = conf.get('test', 'pin')
        d = defer.Deferred()
        def remote_disable_pin_cb():
            d2 = self.sconn.get_pin_status()
            d2.addCallback(lambda status: self.assertEqual(status, 0))
            d3 = self.sconn.enable_pin(pin)
            d3.addCallback(lambda ignored: d.callback(True))

        def remote_disable_pin_eb(failure):
            log.msg("FAILURE RECEIVED %s" % repr(failure))
            log.err(failure)
            d.errback(failure)

        self.remote_obj.Disable(pin,
                                  dbus_interface=PIN_INTFACE,
                                  reply_handler=remote_disable_pin_cb,
                                  error_handler=remote_disable_pin_eb)
        return d

    def test_EnablePIN(self):
        """
        Test that es.warp.Wader.PIN.Enable works
        """
        pin = conf.get('test', 'pin')
        d = defer.Deferred()
        def remote_enable_pin_cb():
            # it seems to have worked
            d2 = self.sconn.get_pin_status()
            d2.addCallback(lambda status: self.assertEqual(status, 1))
            d2.addCallback(lambda _: d.callback(True))

        def remote_enable_pin_eb(exception):
            log.err(exception)
            d.errback(exception)

        self.remote_obj.Enable(pin,
                                  dbus_interface=PIN_INTFACE,
                                  reply_handler=remote_enable_pin_cb,
                                  error_handler=remote_enable_pin_eb)
        return d

    def test_FindContacts(self):
        """
        Test that es.warp.Wader.Contacts.Find result matches find_contacts
        """
        contact = Contact("Eugene", "+43534534")
        d = self.sconn.add_contact(contact)
        def callback(index):
            contact.index = index
            d2 = defer.Deferred()
            def remote_find_contacts_cb(reply):
                self.assertEqual(len(reply), 1)
                reply = list(reply[0])
                self.assertIn(contact.name, reply)
                self.assertIn(contact.number, reply)
                self.assertIn(contact.index, reply)

                d3 = self.sconn.delete_contact(contact.index)
                d3.addCallback(lambda _: d2.callback(True))

            def remote_find_contacts_eb(reply):
                log.err(reply)
                d2.errback(reply)

            self.remote_obj.Find("Euge", dbus_interface=CTS_INTFACE,
                                    reply_handler=remote_find_contacts_cb,
                                    error_handler=remote_find_contacts_eb)
            return d2
        d.addCallback(callback)
        return d

    def test_GetContactByIndex(self):
        """
        Test that es.warp.Wader.Contacts.GetByIndex matches get_contact_by_index
        """
        contact = Contact("Mario", "+312232332")
        d = self.sconn.add_contact(contact)
        def callback(index):
            contact.index = index
            d2 = defer.Deferred()
            def remote_get_contact_by_index_cb(reply):
                reply = list(reply)
                self.assertIn(contact.name, reply)
                self.assertIn(contact.number, reply)
                self.assertIn(contact.index, reply)

                d3 = self.sconn.delete_contact(contact.index)
                d3.addCallback(lambda _: d2.callback(True))

            def remote_get_contact_by_index_eb(reply):
                log.err(reply)
                d2.errback(reply)

            self.remote_obj.GetByIndex(
                                index,
                                dbus_interface=CTS_INTFACE,
                                reply_handler=remote_get_contact_by_index_cb,
                                error_handler=remote_get_contact_by_index_eb)
            return d2
        d.addCallback(callback)
        return d

    def test_ListContacts(self):
        """
        Test that es.warp.Wader.Contacts.List result matches get_contacts
        """
        contact = Contact("Jauma", "+356456445654")
        d = self.sconn.add_contact(contact)
        def callback(index):
            contact.index = index
            d2 = defer.Deferred()
            def get_contacts_cb(contacts):
                def remote_get_contacts_cb(reply):
                    reply = list(reply[0])
                    self.assertIn(contact.name, reply)
                    self.assertIn(contact.number, reply)
                    self.assertIn(contact.index, reply)

                    d3 = self.sconn.delete_contact(contact.index)
                    d3.addCallback(lambda _: d2.callback(True))

                def remote_get_contacts_eb(reply):
                    log.err(reply)
                    d2.errback(reply)

                self.remote_obj.List(dbus_interface=CTS_INTFACE,
                                     reply_handler=remote_get_contacts_cb,
                                     error_handler=remote_get_contacts_eb)

            self.sconn.get_contacts().addCallback(get_contacts_cb)
            return d2

        d.addCallback(callback)
        return d

    def test_GetBands(self):
        """
        Test that es.warp.Wader.Card.GetBands result matches get_bands
        """
        d = self.sconn.get_bands()
        def get_bands_cb(bands):
            d2 = defer.Deferred()
            def remote_get_bands_cb(reply):
                self.assertEqual(reply, bands)
                d2.callback(True)
            def remote_get_bands_eb(reply):
                log.err(reply)
                d2.errback(reply)

            self.remote_obj.GetBands(dbus_interface=CRD_INTFACE,
                                     reply_handler=remote_get_bands_cb,
                                     error_handler=remote_get_bands_eb)
            return d2

        d.addCallback(get_bands_cb)
        return d

    def test_GetCardModel(self):
        """
        Test that es.warp.Wader.Card.GetModel result matches get_card_model
        """
        d = self.sconn.get_card_model()
        def callback(model):
            d2 = defer.Deferred()
            def remote_get_card_model_cb(reply):
                self.assertEqual(model, reply)
                d2.callback(True)
            def remote_get_card_model_eb(reply):
                log.err(reply)
                d2.errback(reply)

            self.remote_obj.GetModel(dbus_interface=CRD_INTFACE,
                                     reply_handler=remote_get_card_model_cb,
                                     error_handler=remote_get_card_model_eb)
            return d2
        d.addCallback(callback)
        return d

    def test_GetCardVersion(self):
        """
        Test that es.warp.Wader.Card.GetVersion result matches get_card_version
        """
        d = self.sconn.get_card_version()
        def callback(version):
            d2 = defer.Deferred()
            def remote_get_card_version_cb(reply):
                self.assertEqual(version, reply)
                d2.callback(True)
            def remote_get_card_version_eb(reply):
                log.err(reply)
                d.errback(reply)

            self.remote_obj.GetVersion(dbus_interface=CRD_INTFACE,
                                   reply_handler=remote_get_card_version_cb,
                                   error_handler=remote_get_card_version_eb)
            return d2
        d.addCallback(callback)
        return d

    def test_GetAvailableCharsets(self):
        """
        Test that es.warp.Wader.Card.GetAvailableCharsets matches
        get_available_charsets
        """
        d = self.sconn.get_available_charset()
        def get_available_charsets_cb(charsets):
            d2 = defer.Deferred()

            def get_av_charsets_cb(reply):
                self.assertEquals(reply, charsets)
                d2.callback(True)

            def get_av_charsets_eb(reply):
                log.err(reply)
                d2.errback(reply)

            self.remote_obj.GetAvailableCharsets(dbus_interface=CRD_INTFACE,
                                                 reply_handler=get_av_charsets_cb,
                                                 error_handler=get_av_charsets_eb)

            return d2

        d.addCallback(get_available_charsets_cb)
        return d

    def test_GetCharset(self):
        """
        Test that es.warp.Wader.Card.GetCharset result matches get_charset
        """
        d = self.sconn.get_charset()
        def callback(charset):
            d2 = defer.Deferred()
            def remote_get_charset_cb(reply):
                self.assertEqual(charset, reply)
                d2.callback(True)
            def remote_get_charset_eb(reply):
                log.err(reply)
                d.errback(reply)

            self.remote_obj.GetCharset(dbus_interface=CRD_INTFACE,
                                    reply_handler=remote_get_charset_cb,
                                    error_handler=remote_get_charset_eb)
            return d2
        d.addCallback(callback)
        return d

    def test_GetIMEI(self):
        """
        Test that es.warp.Wader.Card.GetIMEI result matches get_imei
        """
        d = self.sconn.get_imei()
        def callback(imei):
            d2 = defer.Deferred()
            def remote_get_imei_cb(reply):
                self.assertEqual(imei, reply)
                d2.callback(True)
            def remote_get_imei_eb(reply):
                log.err(reply)
                d.errback(reply)

            self.remote_obj.GetIMEI(dbus_interface=CRD_INTFACE,
                                    reply_handler=remote_get_imei_cb,
                                    error_handler=remote_get_imei_eb)
            return d2
        d.addCallback(callback)
        return d

    def test_GetIMSI(self):
        """
        Test that es.warp.Wader.Card.GetIMSI result matches get_imsi
        """
        d = self.sconn.get_imsi()
        def callback(imsi):
            d2 = defer.Deferred()
            def remote_get_imsi_cb(reply):
                self.assertEqual(imsi, reply)
                d2.callback(True)
            def remote_get_imsi_eb(reply):
                log.err(reply)
                d.errback(reply)

            self.remote_obj.GetIMSI(dbus_interface=CRD_INTFACE,
                                    reply_handler=remote_get_imsi_cb,
                                    error_handler=remote_get_imsi_eb)
            return d2
        d.addCallback(callback)
        return d

    def test_GetManufacturerName(self):
        """
        Test that e.w.V.C.GetManufacturer result matches get_manufacturer_name
        """
        d = self.sconn.get_manufacturer_name()
        def callback(name):
            d2 = defer.Deferred()
            def remote_get_manfname_cb(reply):
                self.assertEqual(name, reply)
                d2.callback(True)
            def remote_get_manfname_eb(reply):
                log.err(reply)
                d.errback(reply)

            self.remote_obj.GetManufacturer(dbus_interface=CRD_INTFACE,
                                    reply_handler=remote_get_manfname_cb,
                                    error_handler=remote_get_manfname_eb)
            return d2
        d.addCallback(callback)
        return d

    def test_GetRegistrationStatus(self):
        """
        Test that e.w.V.N.GetRegistrationStatus matches get_netreg_status
        """
        d = self.sconn.get_netreg_status()
        def callback(status):
            d2 = defer.Deferred()
            def remote_get_netreg_status_cb(reply):
                self.assertEqual(status, reply)
                d2.callback(True)
            def remote_get_netreg_status_eb(reply):
                log.err(reply)
                d.errback(reply)

            self.remote_obj.GetRegistrationStatus(dbus_interface=NET_INTFACE,
                                    reply_handler=remote_get_netreg_status_cb,
                                    error_handler=remote_get_netreg_status_eb)
            return d2
        d.addCallback(callback)
        return d

    def test_GetNetworkInfo(self):
        """
        Test that es.warp.Wader.Network.GetInfo result matches get_network_info
        """
        d = self.sconn.get_network_info()
        def callback(netinfo):
            d2 = defer.Deferred()
            def remote_get_netinfo_cb(reply):
                self.assertEqual(netinfo, reply)
                d2.callback(True)
            def remote_get_netinfo_eb(reply):
                log.err(reply)
                d.errback(reply)

            self.remote_obj.GetInfo(dbus_interface=NET_INTFACE,
                                    reply_handler=remote_get_netinfo_cb,
                                    error_handler=remote_get_netinfo_eb)
            return d2
        d.addCallback(callback)
        return d

    def test_GetNetworkNames(self):
        """
        Test that es.warp.Wader.Network.GetNames matches get_network_names
        """
        raise unittest.SkipTest("Not ready")
        d = self.sconn.get_network_names()
        def callback(netobjs):
            d2 = defer.Deferred()
            def remote_get_netnames_cb(reply):
                for struct in reply:
                    self.assertIn(NetworkOperator(*list(struct)), netobjs)

                d2.callback(True)

            def remote_get_netnames_eb(reply):
                log.err(reply)
                d.errback(reply)

            self.remote_obj.GetNames(dbus_interface=NET_INTFACE,
                                     reply_handler=remote_get_netnames_cb,
                                     error_handler=remote_get_netnames_eb)
            return d2

        d.addCallback(callback)
        return d

    def test_GetPhonebookSize(self):
        """
        Test that e.w.V.C.GetPhonebookSize result matches get_phonebook_size
        """
        d = self.sconn.get_phonebook_size()
        def callback(size):
            d2 = defer.Deferred()
            def remote_get_phonebooksize_cb(reply):
                self.assertEqual(size, reply)
                d2.callback(True)
            def remote_get_phonebooksize_eb(reply):
                log.err(reply)
                d.errback(reply)

            self.remote_obj.GetPhonebookSize(dbus_interface=CRD_INTFACE,
                                    reply_handler=remote_get_phonebooksize_cb,
                                    error_handler=remote_get_phonebooksize_eb)
            return d2

        d.addCallback(callback)
        return d

    def test_GetRoamingIDs(self):
        """
        Test that e.w.V.N.GetRoamingIDs result matches get_roaming_ids
        """
        d = self.sconn.get_roaming_ids()
        def callback(roaming_ids):
            d2 = defer.Deferred()
            def remote_get_smsc_cb(reply):
                rids = [obj.netid for obj in roaming_ids]
                for netid in reply:
                    self.assertIn(netid, rids)

                d2.callback(True)
            def remote_get_smsc_eb(reply):
                log.err(reply)
                d.errback(reply)

            self.remote_obj.GetRoamingIDs(dbus_interface=NET_INTFACE,
                                    reply_handler=remote_get_smsc_cb,
                                    error_handler=remote_get_smsc_eb)
            return d2
        d.addCallback(callback)
        return d

    def test_GetSignal(self):
        """
        Test that e.w.V.N.GetSignal result matches get_signal_level
        """
        d = self.sconn.get_signal_level()
        def callback(rids):
            d2 = defer.Deferred()
            def remote_get_siglevel_cb(reply):
                self.assertEqual(rids, reply)
                d2.callback(True)
            def remote_get_siglevel_eb(reply):
                log.err(reply)
                d.errback(reply)

            self.remote_obj.GetSignal(dbus_interface=NET_INTFACE,
                                      reply_handler=remote_get_siglevel_cb,
                                      error_handler=remote_get_siglevel_eb)
            return d2
        d.addCallback(callback)
        return d

    def test_GetSMSC(self):
        """
        Test that es.warp.Wader.SMS.GetSMSC result matches get_smsc
        """
        d = self.sconn.get_smsc()
        def callback(smsc):
            d2 = defer.Deferred()
            def remote_get_smsc_cb(reply):
                self.assertEqual(smsc, reply)
                d2.callback(True)
            def remote_get_smsc_eb(reply):
                log.err(reply)
                d.errback(reply)

            self.remote_obj.GetSMSC(dbus_interface=SMS_INTFACE,
                                    reply_handler=remote_get_smsc_cb,
                                    error_handler=remote_get_smsc_eb)
            return d2
        d.addCallback(callback)
        return d

    def test_GetFormat(self):
        """
        Test that es.warp.Wader.SMS.GetFormat result matches get_sms_format
        """
        d = self.sconn.get_sms_format()
        def callback(format):
            d2 = defer.Deferred()
            def remote_get_smsc_cb(reply):
                self.assertEqual(format, reply)
                d2.callback(True)
            def remote_get_smsc_eb(reply):
                log.err(reply)
                d.errback(reply)

            self.remote_obj.GetFormat(dbus_interface=SMS_INTFACE,
                                      reply_handler=remote_get_smsc_cb,
                                      error_handler=remote_get_smsc_eb)
            return d2
        d.addCallback(callback)
        return d

    def test_SetCharset(self):
        """
        Test that es.warp.Wader.Card.SetCharset works
        """
        bad_charset = 'IRA'
        old_charset = self.device.sim.charset
        d = defer.Deferred()
        def remote_set_charset_cb():
            def check_and_set_good_charset_cb(charset):
                self.assertEqual(charset, bad_charset)
                d2 = self.sconn.set_charset(old_charset)
                d2.addCallback(lambda _: d.callback(True))
                return d2

            # check that the charset we just set over D-Bus is correct
            d2 = self.sconn.get_charset()
            d2.addCallback(check_and_set_good_charset_cb)
            return d2

        def remote_set_charset_eb(reply):
            log.err(reply)
            d.errback(reply)

        self.remote_obj.SetCharset(bad_charset,
                                dbus_interface=CRD_INTFACE,
                                reply_handler=remote_set_charset_cb,
                                error_handler=remote_set_charset_eb)
        return d

    def test_SetFormat(self):
        """
        Test that es.warp.Wader.SMS.SetFormat works
        """
        def check_and_set_format(format):
            d2 = defer.Deferred()
            def remote_set_format_cb():
                # leave it like we found it
                d3 = self.sconn.set_sms_format(0)
                d3.chainDeferred(d2)

            def remote_set_format_eb(failure):
                log.err(failure)
                d2.errback(failure)

            self.assertEquals(format, 0)
            self.remote_obj.SetFormat(1,
                                      dbus_interface=SMS_INTFACE,
                                      reply_handler=remote_set_format_cb,
                                      error_handler=remote_set_format_eb)

            return d2

        d = self.sconn.get_sms_format()
        d.addCallback(check_and_set_format)
        return d

    def test_SetSMSC(self):
        """
        Test that es.warp.Wader.SMS.SetSMSC works
        """
        badsmsc = '+3453456343'
        d = defer.Deferred()
        def remote_set_smsc_cb():
            def check_and_set_good_smsc_cb(smsc):
                self.assertEqual(smsc, badsmsc)
                d3 = self.sconn.set_smsc(conf.get('test', 'smsc'))
                d3.addCallback(lambda ignored: d.callback(True))
                return d3

            d2 = self.sconn.get_smsc()
            d2.addCallback(check_and_set_good_smsc_cb)
            return d2

        def remote_set_smsc_eb(reply):
            log.err(reply)
            d.errback(reply)

        self.remote_obj.SetSMSC(str(badsmsc),
                                dbus_interface=SMS_INTFACE,
                                reply_handler=remote_set_smsc_cb,
                                error_handler=remote_set_smsc_eb)
        return d
