# -*- coding: utf-8 -*-
# Copyright (C) 2008  Warp Networks, S.L.
# Author:  Pablo Martí­
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""Unittests for the profiles module"""

from twisted.trial import unittest

import dbus.mainloop.glib
gloop = dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)

from twisted.python import log

from wader.common.profiles import ProfileManager
from wader.common.consts import (WADER_PROFILES_OBJECT, WADER_PROFILES_SERVICE,
                                    PRF_INTFACE)

class TestProfiles(unittest.TestCase):
    """Test for profiles module"""

    def setUpClass(self):
        self.manager = ProfileManager()

    def tearDownClass(self):
        self.manager.close()
        del self.manager

    def test_create_profile(self):
        """
        Checks that creating a profile works
        """
        opts = {
            'name'     : 'myprof',
            'username' : 'vodafone',
            'password' : 'vodafone',
            'apn' : 'ac.vodafone.es',
            'connection' : '3GPREF'}

        profile = self.manager.create_profile(opts)
        self.assertIn(profile.opath, self.manager.get_profile_opaths())
        self.manager.delete_profile(profile)

    def test_delete_profile(self):
        """
        Test that deleting a profile works
        """
        opts2 = {
            'name'     : 'alice\'s profile',
            'username' : 'vodafone',
            'password' : 'vodafone',
            'apn' : 'ac.vodafone.es',
            'connection' : '3GONLY'}
        profile = self.manager.create_profile(opts2)
        self.assertIn(profile.opath, self.manager.get_profile_opaths())
        self.manager.delete_profile(profile)
        self.assertNotIn(profile.opath, self.manager.get_profile_opaths())

    def test_get_profile_opaths(self):
        opts3 = {
            'name'     : 'barfoo',
            'username' : 'vodafone',
            'password' : 'vodafone',
            'apn' : 'ac.vodafone.es',
            'connection' : '3GONLY'}
        profile3 = self.manager.create_profile(opts3)

        opts4 = {
            'name'    : 'bob profile',
            'username' : 'vodafone',
            'password' : 'vodafone',
            'apn' : 'ac.vodafone.es',
            'connection' : '3GONLY'}
        profile4 = self.manager.create_profile(opts4)

        self.assertIn(profile3.opath, self.manager.get_profile_opaths())
        self.assertIn(profile4.opath, self.manager.get_profile_opaths())
        self.manager.delete_profile(profile3)
        self.manager.delete_profile(profile4)

    def test_modify_profile(self):
        opts = {
            'name'     : 'juan profile',
            'username' : 'vodafone',
            'password' : 'vodafone',
            'apn' : 'ac.vodafone.es',
            'connection' : '3GONLY'}
        profile = self.manager.create_profile(opts)

        opts2 = {
            'name'     : 'juan profile2',
            'username' : 'vodafoo',
            'password' : 'vodafoo',
            'apn' : 'ac.vodafoo.es',
            'connection' : '3GPREF'}

        profile.patch(opts2)
        self.assertEqual(profile, opts2)
        profile.close()


class TestProfilesDBus(unittest.TestCase):
    """Test for profiles module over DBus"""

    def setUpClass(self):
        bus = dbus.SessionBus()
        self.manager = bus.get_object(WADER_PROFILES_SERVICE,
                                      WADER_PROFILES_OBJECT)

    def tearDownClass(self):
        del self.manager

    def test_CreateProfile(self):
        opts = {
            'name'     : 'dbus profile 1',
            'username' : 'vodafone',
            'password' : 'vodafone',
            'apn' : 'ac.vodafone.es',
            'connection' : '3GONLY'}

        opath = self.manager.CreateProfile(opts,
                                           dbus_interface=PRF_INTFACE)
        self.assertIn(opath,
                      self.manager.GetProfiles(dbus_interface=PRF_INTFACE))
        self.manager.DeleteProfile(opath, dbus_interface=PRF_INTFACE)
        self.assertNotIn(opath,
                      self.manager.GetProfiles(dbus_interface=PRF_INTFACE))

    def test_DeleteProfile(self):
        opts = {
            'name'     : 'dbus profile 2',
            'username' : 'vodafone',
            'password' : 'vodafone',
            'apn' : 'ac.vodafone.es',
            'connection' : '3GONLY'}

        opts2 = {
            'name'     : 'dbus profile 3',
            'username' : 'three',
            'password' : 'three',
            'apn' : 'apn.three.co.uk',
            'connection' : '3GONLY'}

        opath = self.manager.CreateProfile(opts,
                                           dbus_interface=PRF_INTFACE)
        opath2 = self.manager.CreateProfile(opts2,
                                           dbus_interface=PRF_INTFACE)
        self.assertIn(opath,
                      self.manager.GetProfiles(dbus_interface=PRF_INTFACE))
        self.assertIn(opath2,
                      self.manager.GetProfiles(dbus_interface=PRF_INTFACE))
        self.manager.DeleteProfile(opath, dbus_interface=PRF_INTFACE)
        self.manager.DeleteProfile(opath2, dbus_interface=PRF_INTFACE)
        self.assertNotIn(opath,
                      self.manager.GetProfiles(dbus_interface=PRF_INTFACE))
        self.assertNotIn(opath2,
                      self.manager.GetProfiles(dbus_interface=PRF_INTFACE))

    def test_ModifyProfile(self):
        opts = {
            'name'     : 'dbus profile 4',
            'username' : 'vodafone',
            'password' : 'vodafone',
            'apn' : 'ac.vodafone.es',
            'connection' : '3GONLY'}
        opath = self.manager.CreateProfile(opts,
                                           dbus_interface=PRF_INTFACE)
        opts2 = {
            'name'     : 'dbus profile 5',
            'username' : 'three',
            'password' : 'three',
            'apn' : 'apn.three.co.uk',
            'connection' : '3GONLY'}
        log.msg("Modifying profile %s" % opath)
        try:
            self.manager.ModifyProfile(opath, opts2,
                                        dbus_interface=PRF_INTFACE)
        except:
            log.err()
        bus = dbus.SessionBus()
        log.msg("Getting a reference to %s" % opath)
        try:
            obj = bus.get_object(WADER_PROFILES_SERVICE, opath)
        except:
            log.err()
        log.msg("Comparing %s with %s" % (opath, opts2))
        for k, v in opts2.iteritems():
            self.assertEqual(v, obj.Get(k))

        self.manager.DeleteProfile(opath, dbus_interface=PRF_INTFACE)
